import networkx as nx
import pandas as pd
from IPython.display import Image, display

import scphylo as scp
from scphylo.pl._helper import (
    _add_barplot,
    _add_chromplot,
    _clonal_cell_mutation_list,
    _get_tree,
    _newick_info2_mutation_list,
)


def clonal_tree(
    tree,
    muts_as_number=False,
    cells_as_number=False,
    show_id=False,
    cell_info=None,
    output_file=None,
    color_attr=None,
    dpi=150,
):
    """Draw the tree in clonal format.

    This functions plots the tree in which edges are mutations and nodes are cells.

    Parameters
    ----------
    tree : :class:`networkx.DiGraph`
        The input tree.
    muts_as_number : :obj:`bool`, optional
        Change the mutation list to a number at edges, by default False
    cells_as_number : :obj:`bool`, optional
        Change the cell list to a number at edges, by default False
    show_id : :obj:`bool`, optional
        Whether to show IDs of nodes and edges or not, by default True
    cell_info : :class:`pandas.DataFrame`, optional
        Information of cells for coloring the nodes by a pie chart, by default None
    output_file : :obj:`str`, optional
        Path to a file for saving the tree in, by default None
    color_attr : :obj:`str`, optional
        Attributes in the `cell_info` dataframe for coloring the nodes, by default None
    dpi : :obj:`int`, optional
        Resolution of rendered figures – this influences the size of
        figures in notebooks, by default 150

    Returns
    -------
    :obj:`None`
    """
    _, graphviz_is_not_imporeted = scp.ul.import_graphviz()
    if graphviz_is_not_imporeted:
        scp.logg.error("Unable to import a package!")

    tc = tree.copy()
    root = scp.ul.root_id(tree)
    tc.nodes[root]["label"] = "root"
    tc.nodes[root]["fontname"] = "Helvetica"
    tc.nodes[root]["style"] = "rounded"
    tc.nodes[root]["shape"] = "box"
    tc.nodes[root]["margin"] = 0.05
    tc.nodes[root]["pad"] = 0
    tc.nodes[root]["width"] = 0
    tc.nodes[root]["height"] = 0

    if muts_as_number:
        for u, v, label in tc.edges.data("label"):
            if label == "":
                ll = []
            else:
                ll = label.split(tc.graph["splitter_mut"])
            tc.add_edge(u, v, label=f"  {len(ll)}  ")

    if cells_as_number:
        for n in tc.nodes:
            if n != root:
                ll = tc.nodes[n]["label"].split(tc.graph["splitter_cell"])
                if "––" in tc.nodes[n]["label"]:
                    tc.nodes[n]["label"] = "0"
                else:
                    tc.nodes[n]["label"] = f"{len(ll)}"

    if cell_info is not None:
        tc.nodes[root]["label"] = tree.graph["splitter_cell"].join(
            tc.graph["normal_cells"]
        )
        mapping = cell_info[color_attr].to_dict()
        for node in tc:
            num = 0
            paths = nx.shortest_path(tc, source=root, target=node)
            for i in range(len(paths) - 1):
                x = paths[i]
                y = paths[i + 1]
                num += len(tc[x][y]["label"].split(tc.graph["splitter_mut"]))
            try:
                freq = [
                    mapping[x]
                    for x in tc.nodes[node]["label"].split(tc.graph["splitter_cell"])
                ]
            except KeyError:
                freq = ["#FFFFFF"]
            freq = pd.DataFrame(freq)[0].value_counts(normalize=True)
            fillcolor = ""
            for index, value in freq.items():
                fillcolor += f"{index};{value}:"
            tc.nodes[node]["fontsize"] = 14
            tc.nodes[node]["shape"] = "circle"
            tc.nodes[node]["fontname"] = "Helvetica"
            tc.nodes[node]["style"] = "wedged"
            tc.nodes[node]["margin"] = 0.05
            tc.nodes[node]["pad"] = 0
            if "––" in tc.nodes[node]["label"]:
                tc.nodes[node]["width"] = 0
                tc.nodes[node]["height"] = 0
            else:
                tc.nodes[node]["width"] = 0.8
                tc.nodes[node]["height"] = 0.8
            tc.nodes[node]["fillcolor"] = fillcolor  # "red;0.3:green;0.6:orange"
            tc.nodes[node]["label"] = len(
                tc.nodes[node]["label"].split(tc.graph["splitter_cell"])
            )
            tc.nodes[node]["fontcolor"] = "white"
            tc.nodes[node]["color"] = "gray"

    if show_id:
        for u, v, label in tc.edges.data("label"):
            tc.add_edge(u, v, label=label + f"\n[{v}]")
            tc.nodes[v]["label"] = tc.nodes[v]["label"] + f"\n[{v}]"

    tc.graph["graph"] = {"fontname": "Helvetica"}
    tc.graph["node"] = {"fontname": "Helvetica", "fontsize": 14}
    tc.graph["edge"] = {"fontname": "Helvetica", "fontsize": 14}

    tree.graph["type"] = "clonal"
    cell_list, mutation_list = _clonal_cell_mutation_list(tree)
    tree.graph["cell_list"] = cell_list
    tree.graph["mutation_list"] = mutation_list

    if output_file is not None:
        scp.io.to_png(tc, output_file, dpi)

    return display(
        Image(nx.drawing.nx_pydot.to_pydot(tc).create_png(), embed=True, retina=True)
    )


def dendro_tree(
    tree,
    width=1200,
    height=500,
    dpi=300,
    cell_info=None,
    label_color="group_color",
    line_size=0.4,
    tiplab_size=2.5,
    inner_node_type="nmuts",
    inner_node_size=2,
    distance_labels_to_bottom=4,
    annotation=None,
    output_file=None,
):
    """Draw the tree in dendro fromat.

    Parameters
    ----------
    tree : :class:`networkx.DiGraph`
        The input tree.
    width : :obj:`int`, optional
        Width of the figure, by default 8
    height : :obj:`int`, optional
        Height of the figure, by default 3
    dpi : :obj:`int`, optional
        The resolution, by default 300
    cell_info : :class:`pandas.DataFrame`, optional
        Information about cells such as color,
        expression values of genes and etc, by default None
    label_color : :obj:`str`, optional
        The column name in which colors of cells are stored
        in the dataframe provided as `cell_info`, by default "group_color"
    line_size : :obj:`float`, optional
        Line size of the tree, by default 0.4
    tiplab_size : :obj:`float`, optional
        Cell name size in the tree, by default 2.5
    inner_node_type : :obj:`str`, optional
        The format of the inner nodes (i.e. mutations), by default "nmuts"
        Values are:

            - `nmuts`: only number of mutations
            - `nodeid`: only node id
            - `both`: both number of mutations and node id
    inner_node_size : :obj:`int`, optional
        Size of the inner nodes (i.e. mutations), by default 2
    distance_labels_to_bottom : :obj:`int`, optional
        Distance cell names to the bottom of the figure, by default 4
    annotation : :obj:`list`, optional
        List of gene names provided in the column dataframe of ``cell_info``
        in to be annotated in the bottom of the tree, by default []
    output_file : :obj:`str`, optional
        Path to a file for saving the tree in, by default None

    Returns
    -------
    :obj:`None`

    Note
    ----
    The cell names in the tree must be identical to the index of `cell_info`
    dataframe if it was provided.
    """
    if annotation is None:
        annotation = []

    if inner_node_type.lower() not in ["nmuts", "nodeid", "both"]:
        scp.logg.error("Wrong `inner_node_type` choice!")

    ggtree, ggtree_is_not_imported = scp.ul.import_rpy2(
        "ggtree",
        "devtools::install_github(c('YuLab-SMU/ggtree','xiangpin/ggtreeExtra'"
        + ",'YuLab-SMU/aplot'))\ninstall.packages('cowplot')\n",
    )
    if ggtree_is_not_imported:
        scp.logg.error("Unable to import a package!")

    import rpy2.robjects as ro
    from rpy2.robjects import pandas2ri
    from rpy2.robjects.lib import grdevices
    from rpy2.robjects.packages import importr

    newick, info2, mutation_list = _newick_info2_mutation_list(tree)
    tree.graph["mutation_list"] = mutation_list.set_index("index")
    tree.graph["newick"] = newick
    tree.graph["type"] = "dendro"

    importr("ggplot2")
    importr("cowplot")
    importr("ggtree")
    importr("ape")
    importr("aplot")

    with ro.conversion.localconverter(ro.default_converter + pandas2ri.converter):
        if cell_info is not None:
            if "group_color" not in cell_info.columns:
                cell_info["group_color"] = tree.graph["data"].shape[0] * ["#000000"]
            cell_info_r = ro.conversion.py2rpy(cell_info.reset_index())
        else:
            cell_info_r = ro.conversion.py2rpy(
                pd.DataFrame.from_dict(
                    {
                        "x": tree.graph["data"].index,
                        "group_color": tree.graph["data"].shape[0] * ["#000000"],
                    }
                )
            )
        ro.globalenv["info1"] = cell_info_r
        info2_r = ro.conversion.py2rpy(info2)
        ro.globalenv["info2"] = info2_r

    cmd = _get_tree(
        newick,
        line_size,
        label_color,
        tiplab_size,
        inner_node_type,
        inner_node_size,
        distance_labels_to_bottom,
    )

    for i, ann in enumerate(annotation):
        if i == 0:
            cmd += "info1$id_index <- p$data[match(info1[,1], p$data$label),]$y"
        if ann[0] == "bar":
            cmd += _add_barplot(ann[1], ann[2], ann[3])
        elif ann == "chrom":
            cmd += _add_chromplot()
        elif ann == "imputation":
            for imp in [
                ("f_0_1", "f_0_1_color"),
                ("f_1_0", "f_1_0_color"),
                ("f_3_0", "f_3_0_color"),
                ("f_3_1", "f_3_1_color"),
            ]:
                cmd += _add_barplot(imp[0], imp[1])

    with grdevices.render_to_bytesio(
        grdevices.png, width=width, height=height, res=dpi
    ) as image:
        p = ro.r(cmd)
        ro.r.show(p)
        if output_file is not None:
            ro.r.ggsave(
                plot=p,
                filename=output_file,
                width=width / dpi,
                height=height / dpi,
                units="in",
                dpi=dpi,
                limitsize=False,
            )
    return display(Image(image.getvalue(), embed=True, retina=True))


def newick_tree(
    adata,
    tree_attr="tree",
    leaf_attr=None,
    width=2000,
    height=800,
    dpi=300,
    tippoint_size=0.8,
    tiplab_size=0.9,
    show_tiplab=False,
):
    """[summary].

    Parameters
    ----------
    adata : [type]
        [description]
    leaf_attr : [type]
        [description]
    tree_attr : [type]
        [description]
    width : int, optional
        [description], by default 2000
    height : int, optional
        [description], by default 800
    dpi : int, optional
        [description], by default 300
    tippoint_size : float, optional
        [description], by default 0.8
    tiplab_size : float, optional
        [description], by default 0.9
    show_tiplab : bool, optional
        [description], by default False

    Returns
    -------
    [type]
        [description]
    """
    ggtree, ggtree_is_not_imported = scp.ul.import_rpy2(
        "ggtree",
        "devtools::install_github(c('YuLab-SMU/ggtree','xiangpin/ggtreeExtra'"
        + ",'YuLab-SMU/aplot'))\ninstall.packages('cowplot')\n",
    )
    if ggtree_is_not_imported:
        scp.logg.error("Unable to import a package!")

    import rpy2.robjects as ro
    from rpy2.robjects import pandas2ri
    from rpy2.robjects.lib import grdevices
    from rpy2.robjects.packages import importr

    importr("ggplot2")

    if leaf_attr is not None:
        df = adata.obs[[leaf_attr]]
    else:
        df = adata.obs.copy()
        df["color"] = "#000000"
        df = df[["color"]]
        leaf_attr = "color"
    with ro.conversion.localconverter(ro.default_converter + pandas2ri.converter):
        info_color_r = ro.conversion.py2rpy(df.reset_index())
        ro.globalenv["info"] = info_color_r

    cmd = f"""
    tree <- read.tree(text='{str(adata.uns[tree_attr]).strip()}')
    p <- ggtree(tree, layout='dendrogram', size=0.2, color='gray')
    # p <- p %>% rotate(191) %>% rotate(218)
    # p <- p + geom_text(aes(label=node), size=1)
    p <- p %<+% info + geom_tippoint(aes(color={leaf_attr}), size={tippoint_size})
    """
    if show_tiplab:
        cmd += f"""
        p <- p + geom_tiplab(aes(color={leaf_attr}), angle=90,
                                 size={tiplab_size}, hjust=1.1)
        """
    cmd += f"""
    p <- p + scale_color_identity(guide='none')
    # ggsave('/data/frashidi/{leaf_attr}.pdf', p, width=7.2, height=2,
    #        units='in', limitsize=FALSE, dpi=300)
    p
    """

    with grdevices.render_to_bytesio(
        grdevices.png, width=width, height=height, res=dpi
    ) as image:
        p = ro.r(cmd)
        ro.r.show(p)
    return display(Image(image.getvalue(), embed=True, retina=True))


def networkx_tree(tree, n_attr="label", e_attr="label"):
    """[summary].

    Parameters
    ----------
    tree : [type]
        [description]
    n_attr : str, optional
        [description], by default "label"
    e_attr : str, optional
        [description], by default "label"
    """
    graph2 = tree.copy()
    if n_attr == "id":
        for n in graph2.nodes():
            if graph2.in_degree(n) == 1 and graph2.out_degree(n) >= 1:
                graph2.nodes[n]["shape"] = "circle"
                graph2.nodes[n]["margin"] = 0
                graph2.nodes[n]["label"] = n
    else:
        for n in graph2.nodes():
            if n_attr in graph2.nodes[n]:
                graph2.nodes[n]["label"] = graph2.nodes[n][n_attr]
    for e in graph2.edges():
        if e_attr in graph2.edges[e]:
            graph2.edges[e]["label"] = graph2.edges[e][e_attr]
    display(Image(nx.drawing.nx_pydot.to_pydot(graph2).create_png(), retina=True))



import subprocess

import click

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run VarScan2.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "normal",
    required=True,
    type=str,
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default="0-10:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="50",
    type=str,
    show_default=True,
    help="Memory.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def varscan2(outdir, normal, ref, time, mem, afterok):
    """Run VarScan2.

    scphylo caller varscan2 /path/to/in/dir normal_name hg19|hg38|mm10

    BAM files (*.markdup_bqsr.bam) --> VCF files (*.varscan2.vcf)
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def step1(afterok):
        def get_command(sample):
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['samtools']}"])
            cmds += cmd(
                [
                    "samtools",
                    "mpileup",
                    "-B",
                    f"-f {config['ref']}",
                    f"{outdir}/{sample}.markdup_bqsr.bam",
                    f"> {outdir}/{sample}.markdup_bqsr.pileup",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds_tmp = scp.ul.get_samples_df(outdir, normal=None)
        df_cmds_tmp["cmd"] = df_cmds_tmp.apply(
            lambda x: get_command(x["sample"]), axis=1
        )
        cmdmain = write_cmds_get_main(
            df_cmds_tmp,
            "varscan2-1of2",
            time,
            mem,
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step2(afterok):
        def get_command(sample):
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['varscan']}"])
            cmds += cmd([f"mkdir -p {outdir}/{sample}.varscan2"])
            cmds += cmd(
                [
                    "varscan",
                    "somatic",
                    f"{outdir}/{normal}.markdup_bqsr.pileup",
                    f"{outdir}/{sample}.markdup_bqsr.pileup",
                    f"--output-snp {outdir}/{sample}.varscan2/snp.vcf",
                    f"--output-indel {outdir}/{sample}.varscan2/indel.vcf",
                    "--output-vcf",
                ]
            )
            cmds += cmd([f"module load {scp.settings.tools['bcftools']}"])
            cmds += cmd(
                [
                    "bgzip",
                    f"{outdir}/{sample}.varscan2/snp.vcf",
                ]
            )
            cmds += cmd(
                [
                    "bgzip",
                    f"{outdir}/{sample}.varscan2/indel.vcf",
                ]
            )
            cmds += cmd(
                [
                    "bcftools",
                    "index",
                    f"{outdir}/{sample}.varscan2/snp.vcf.gz",
                ]
            )
            cmds += cmd(
                [
                    "bcftools",
                    "index",
                    f"{outdir}/{sample}.varscan2/indel.vcf.gz",
                ]
            )
            cmds += cmd(
                [
                    "bcftools",
                    "concat",
                    f"{outdir}/{sample}.varscan2/snp.vcf.gz",
                    f"{outdir}/{sample}.varscan2/indel.vcf.gz",
                    f"--output {outdir}/{sample}.varscan2.tmp.vcf",
                    "--output-type v",
                    "--allow-overlaps",
                ]
            )

            cmds += cmd(
                [
                    "varscan",
                    "processSomatic",
                    f"{outdir}/{sample}.varscan2.tmp.vcf",
                ]
            )
            cmds += cmd(
                [
                    "bgzip",
                    f"{outdir}/{sample}.varscan2.tmp.Somatic.vcf",
                ]
            )
            cmds += cmd(
                [
                    "bgzip",
                    f"{outdir}/{sample}.varscan2.tmp.LOH.vcf",
                ]
            )
            cmds += cmd(
                [
                    "bcftools",
                    "index",
                    f"{outdir}/{sample}.varscan2.tmp.Somatic.vcf.gz",
                ]
            )
            cmds += cmd(
                [
                    "bcftools",
                    "index",
                    f"{outdir}/{sample}.varscan2.tmp.LOH.vcf.gz",
                ]
            )
            cmds += cmd(
                [
                    "bcftools",
                    "concat",
                    f"{outdir}/{sample}.varscan2.tmp.Somatic.vcf.gz",
                    f"{outdir}/{sample}.varscan2.tmp.LOH.vcf.gz",
                    f"--output {outdir}/{sample}.varscan2.vcf",
                    "--output-type v",
                    "--allow-overlaps",
                ]
            )
            cmds += cmd(
                [
                    "bcftools",
                    "sort",
                    f"{outdir}/{sample}.varscan2.vcf",
                    f"--output-file {outdir}/{sample}.varscan2.vcf",
                    "--output-type v",
                ]
            )

            cmds += cmd(
                [
                    f"rm -rf {outdir}/{sample}.varscan2",
                    f"rm -rf {outdir}/{sample}.markdup_bqsr.pileup",
                    f"rm -rf {outdir}/{sample}.varscan2.tmp.Germline.hc.vcf",
                    f"rm -rf {outdir}/{sample}.varscan2.tmp.Germline.vcf",
                    f"rm -rf {outdir}/{sample}.varscan2.tmp.Somatic.hc.vcf",
                    f"rm -rf {outdir}/{sample}.varscan2.tmp.LOH.hc.vcf",
                    f"rm -rf {outdir}/{sample}.varscan2.tmp.vcf",
                    f"rm -rf {outdir}/{sample}.varscan2.tmp.Somatic.vcf.gz",
                    f"rm -rf {outdir}/{sample}.varscan2.tmp.Somatic.vcf.gz.csi",
                    f"rm -rf {outdir}/{sample}.varscan2.tmp.LOH.vcf.gz",
                    f"rm -rf {outdir}/{sample}.varscan2.tmp.LOH.vcf.gz.csi",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds_tmp = scp.ul.get_samples_df(outdir, normal)
        df_cmds_tmp["cmd"] = df_cmds_tmp.apply(
            lambda x: get_command(x["sample"]), axis=1
        )
        cmdmain = write_cmds_get_main(
            df_cmds_tmp,
            "varscan2-2of2",
            time,
            mem,
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    cmd1 = step1(afterok)
    code1 = subprocess.getoutput(cmd1)
    cmd2 = step2(code1)
    code2 = subprocess.getoutput(cmd2)
    scp.logg.info(code2)

    return None



import warnings

import numba
import numpy as np
import pandas as pd
import scipy as sp
from scipy.cluster.hierarchy import fcluster, linkage
from sklearn.metrics import pairwise_distances

import scphylo as scp
from scphylo.external._betabinom import pmf_BetaBinomial


@numba.jit(nopython=True)
def _l1_ignore_na(a, b):
    a[a == 3] = np.nan
    b[b == 3] = np.nan
    return np.nanmean(np.abs(a - b))


def _l1_ignore_na_wrapper(x, y, **kwargs):
    return _l1_ignore_na(x, y)


def dist_l1_ignore_na(I_mtr, n_jobs=1):
    dist = pairwise_distances(
        I_mtr, metric=_l1_ignore_na_wrapper, force_all_finite="allow-nan", n_jobs=n_jobs
    )
    np.fill_diagonal(dist, 0)
    return dist


# https://gist.github.com/FedericoV/0e7d6d8c8794a99a7a42
@numba.jit(nopython=True)
def _cosine_ignore_na(u, v):
    m = u.shape[0]
    udotv = 0
    u_norm = 0
    v_norm = 0
    for i in range(m):
        if (np.isnan(u[i])) or (np.isnan(v[i])):
            continue
        udotv += u[i] * v[i]
        u_norm += u[i] * u[i]
        v_norm += v[i] * v[i]
    u_norm = np.sqrt(u_norm)
    v_norm = np.sqrt(v_norm)
    if (u_norm == 0) or (v_norm == 0):
        ratio = 1.0
    else:
        ratio = 1 - udotv / (u_norm * v_norm)
    if ratio < 0:
        return 0
    return ratio


def _cosine_ignore_na_wrapper(x, y, **kwargs):
    return _cosine_ignore_na(x, y)


def dist_cosine_ignore_na(I_mtr, n_jobs=1):
    dist = pairwise_distances(
        I_mtr,
        metric=_cosine_ignore_na_wrapper,
        force_all_finite="allow-nan",
        n_jobs=n_jobs,
    )
    np.fill_diagonal(dist, 0)
    return dist


def _dist_dendro(T, V, I_mtr):
    warnings.filterwarnings("ignore")
    PROB_SEQ_ERROR = 0.001

    def logSum_1(x, y):
        big = np.copy(x)
        big[x < y] = y[x < y]
        small = np.copy(x)
        small[x >= y] = y[x >= y]
        tmp = big + np.log(1 + np.exp(small - big))
        # tmp[np.bitwise_and(x==-np.inf, y==-np.inf)] = -np.inf
        # tmp[np.bitwise_and(x==np.inf, y==np.inf)] = np.inf
        return tmp

    D = np.divide(V, T)

    Mu = np.nanmean(D, axis=0)
    Var = np.nanvar(D, axis=0, ddof=1)
    a = ((1 - Mu) * Mu / Var - 1) * Mu
    b = ((1 - Mu) * Mu / Var - 1) * (1 - Mu)
    bad_muts = (
        (a <= 0) | (b <= 0) | np.isnan(a) | np.isnan(b) | np.isinf(a) | np.isinf(b)
    )
    V = V[:, ~bad_muts]
    T = T[:, ~bad_muts]
    # D = D[:, ~bad_muts]
    I_mtr = I_mtr[:, ~bad_muts]
    a = a[~bad_muts]
    b = b[~bad_muts]

    lPz0 = np.zeros(T.shape, dtype=np.float64)
    lPz1 = np.zeros(T.shape, dtype=np.float64)
    for i in range(T.shape[0]):
        for j in range(T.shape[1]):
            if T[i, j] != 0:
                lPz0[i, j] = np.log(
                    sp.stats.binom.pmf(V[i, j], T[i, j], PROB_SEQ_ERROR)
                )
                lPz1[i, j] = np.log(pmf_BetaBinomial(V[i, j], T[i, j], a[j], b[j]))

    Pg = np.sum(I_mtr == 1, axis=0) / I_mtr.shape[0]
    lPg = np.log(Pg)
    l1Pg = np.log(1 - Pg)
    lupiall = logSum_1(lPz0 + l1Pg, lPz1 + lPg)

    dist = np.zeros((T.shape[0], T.shape[0]), dtype=np.float64)
    for i in range(T.shape[0]):
        ldowni = logSum_1(lPz0[i, :] + lPz0 + l1Pg, lPz1[i, :] + lPz1 + lPg)
        lupi = logSum_1(lupiall[i, :] + lupiall, ldowni)
        dist[i, :] = np.sum(lupi - ldowni, axis=1)

    dist = dist - np.min(dist) + 1
    return dist, bad_muts


def dist_dendro(adata):
    T = adata.layers["total"]
    V = adata.layers["mutant"]
    G = adata.layers["genotype"]
    G[(G == 1) | (G == 3)] = 1
    G[G == 2] = 0
    dist, bad_muts = _dist_dendro(T, V, G)
    scp.pp.remove_mut_by_list(adata, bad_muts)
    scp.logg.info(f"{sum(bad_muts)} mutations filtered")
    return dist


def hclustering(df, metric="l1", method="ward", return_dist=False):
    """Hierarchical clustering.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        The genotype matrix.
    metric: :obj:`str`, optional
        The metric option. Can be:

            - `l1`
            - `cosine`
    method : :obj:`str`, optional
        The method for the hierarchical clustering, by default "ward"

    Returns
    -------
    :obj:`dict`
        A dictionary in which keys are the number of clusters and
        values are the cluster labels for each item.
    """
    if metric == "l1":
        dist = dist_l1_ignore_na(df.values)
    elif metric == "cosine":
        dist = dist_cosine_ignore_na(df.values)
    else:
        scp.logg.error("Wroing `metric` choice!")
    clust = linkage(dist[np.triu_indices(dist.shape[0], 1)], method=method)
    clusters = {}

    for i in range(2, dist.shape[0]):
        fc = fcluster(clust, i, criterion="maxclust")
        clusters[i] = pd.Series(fc, index=df.index)

    if return_dist:
        return dist

    return clusters



import glob
import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run VariantFiltration.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default="1-00:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="90",
    type=str,
    show_default=True,
    help="Memory.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def varfilter(outdir, ref, time, mem, afterok):
    """Run varfilter.

    scphylo caller filter /path/to/vcf/file hg19|hg38|mm10

    VCF file --> VCF.filtered file
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def get_command(sample):
        cmds = ""
        cmds += cmd([f"module load {scp.settings.tools['gatk']}"])
        cmds += cmd(
            [
                f'gatk --java-options "-Xmx{int(mem)-10}g"',
                "IndexFeatureFile",
                f"--input {outdir}/{sample}.vcf",
            ]
        )
        cmds += cmd(
            [
                f'gatk --java-options "-Xmx{int(mem)-10}g"',
                "VariantFiltration",
                f"--reference {config['ref']}",
                f"--variant {outdir}/{sample}.vcf",
                f"--output {outdir}/{sample}.tmp.filtered.vcf",
                "--cluster-window-size 35",
                "--cluster-size 3",
                '--filter-name "FS"',
                '--filter-expression "FS > 30.0"',
                '--filter-name "QD"',
                '--filter-expression "QD < 2.0"',
            ]
        )
        cmds += cmd(
            [
                f'gatk --java-options "-Xmx{int(mem)-10}g"',
                "SelectVariants",
                f"--variant {outdir}/{sample}.tmp.filtered.vcf",
                f"--output {outdir}/{sample}.filtered.vcf",
                "--exclude-filtered",
            ]
        )
        cmds += cmd(
            [
                "rm -rf",
                f"{outdir}/{sample}.tmp.filtered.vcf",
            ]
        )
        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    files = glob.glob(f"{outdir}/*.vcf")
    temp = []
    for file in files:
        file = file.split("/")[-1].replace(".vcf", "")
        temp.append({"sample": file})
    df_cmds = pd.DataFrame(temp)
    df_cmds["cmd"] = df_cmds.apply(lambda x: get_command(x["sample"]), axis=1)

    cmdmain = write_cmds_get_main(
        df_cmds,
        "varfilter",
        time,
        mem,
        None,
        1,
        scp.settings.tools["email"],
        f"{outdir}/_tmp",
        afterok,
    )
    code = subprocess.getoutput(cmdmain)
    scp.logg.info(code)

    return None



import os

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run vartrix.")
@click.argument(
    "bam",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "barcodes",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "vcf",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default="2-00:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="150",
    type=str,
    show_default=True,
    help="Memory.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def vartrix(bam, barcodes, vcf, ref, time, mem, afterok):
    """Run vartrix.

    scphylo caller vartrix /path/bam /path/barcodes /path/vcf hg19|hg38|mm10|m10x

    * --> MTX matrices
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    elif ref == "m10x":
        config = scp.settings.m10x
    else:
        config = None

    outdir, _ = scp.ul.dir_base(bam)
    scp.ul.mkdir(f"{outdir}/_vartrix")

    def cmds():
        cmds = ""
        cmds += cmd([f"module load {scp.settings.tools['vartrix']}"])
        cmds += cmd([f"cp {barcodes} {outdir}/_vartrix/"])
        cmds += cmd(
            [
                "vartrix",
                f"--bam {bam}",
                f"--cell-barcodes {barcodes}",
                f"--vcf {vcf}",
                f"--fasta {config['ref']}",
                "--threads $SLURM_CPUS_PER_TASK",
                "--scoring-method coverage",  # coverage, consensus, alt_frac
                "--umi",
                "--no-duplicates",
                f"--out-matrix {outdir}/_vartrix/alt.mtx",
                f"--ref-matrix {outdir}/_vartrix/ref.mtx",
                f"--out-variants {outdir}/_vartrix/var_names.txt",
                # f"--out-matrix {outdir}/_vartrix/gen.mtx",
            ]
        )
        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    df_cmds = pd.DataFrame()
    df_cmds["cmd"] = [cmds()]

    cmdmain = write_cmds_get_main(
        df_cmds,
        "vartrix",
        time,
        mem,
        None,
        16,
        scp.settings.tools["email"],
        f"{outdir}/_tmp",
        afterok,
    )
    os.system(cmdmain)

    return None



import os
import time

import scphylo as scp


def sciphi(df_input):
    # TODO: implement
    executable = scp.ul.executable("sciphi", "SCIPhI")

    scp.logg.info("running SCIPhI with")

    # tmpdir = scp.ul.tmpdirsys(suffix=".sciphi")
    tmpdir = "test"
    scp.ul.cleanup(tmpdir)
    scp.ul.mkdir(tmpdir)

    matrix_I = df_input.values
    with open(f"{tmpdir}/sciphi.mpileup", "w") as fout:
        for j in range(matrix_I.shape[1]):
            line = f"seq1\t{(j+1)*100}\tA"
            r = q = ""
            for i in range(matrix_I.shape[0]):
                if matrix_I[i, j] == 0:
                    r = "."
                elif matrix_I[i, j] == 1:
                    r = "T"
                elif matrix_I[i, j] == 3:
                    r = "N"
                q = "<"
                line = f"{line}\t1\t{r}\t{q}"
            fout.write(line + "\n")
    with open(f"{tmpdir}/sciphi.cellnames", "w") as fout:
        for i in range(matrix_I.shape[0]):
            fout.write(f"{df_input.index[i]}\tCT\n")

    cmd = (
        f"{executable} "
        f"-o {tmpdir}/out "
        f"--in {tmpdir}/sciphi.cellnames "
        "--seed 42 "
        f"{tmpdir}/sciphi.mpileup "
        f"> {tmpdir}/sciphi.log"
    )

    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time
    type(running_time)

    return None



from collections import Counter, defaultdict
from functools import partial
from itertools import combinations
from multiprocessing import Pool, set_start_method

import networkx as nx
import numpy as np

__author__ = "Simone Ciccolella"
__date__ = "9/24/21"


class Tree:
    def __init__(self, T, label_to_nodes, node_to_labels, label_set):
        """Tree class.

        Parameters
        ----------
        T : [type]
            [description]
        label_to_nodes : [type]
            [description]
        node_to_labels : [type]
            [description]
        label_set : [type]
            [description]
        """
        self.T = T
        self.label_to_nodes = label_to_nodes
        self.node_to_labels = node_to_labels
        self.label_set = label_set

        self.LCA = LCA(self.T, self.label_to_nodes, self.node_to_labels)


class LCA:
    def __init__(self, T, T_label_to_node, T_node_to_labels):
        """LCA class.

        Parameters
        ----------
        T : [type]
            [description]
        T_label_to_node : [type]
            [description]
        T_node_to_labels : [type]
            [description]
        """
        self.LCA_dict = {}
        self.LCA_label_dict = T_label_to_node
        self.LCA_node2lbl = T_node_to_labels
        for lca in nx.tree_all_pairs_lowest_common_ancestor(T):
            self.LCA_dict[(lca[0][0], lca[0][1])] = lca[1]

    def lca_nodes(self, node1, node2):
        try:
            return self.LCA_dict[node1, node2]
        except KeyError:
            return self.LCA_dict[node2, node1]

    def lca_labels(self, label1, label2):
        nodes1 = self.LCA_label_dict[label1]
        nodes2 = self.LCA_label_dict[label2]

        lca_mset = Counter()
        for n1 in nodes1:
            for n2 in nodes2:
                lca_mset.update(self.lca_nodes(n1, n2))

        return lca_mset

    def label_to_node(self, label):
        return self.LCA_label_dict[label]

    def node_to_labels(self, node):
        return self.LCA_node2lbl[node]

    def __str__(self):
        """Return."""
        return str(self.LCA_dict)


class ExtValue:
    def __init__(self):
        """Extvalue class."""
        self.counter = 1
        self.nodes = defaultdict(set)

    def get(self, node):
        if node not in self.nodes:
            self.nodes[node].add(f"EXT{self.counter}")
            self.counter += 1
        return self.nodes[node]


def sigmoid(x, mult=10.0):
    if x == 0:
        return 0
    if x == 1:
        return 1
    return 1 / (1 + np.exp(-mult * (x - 0.5)))


def intersect_mset_card(list_lca1, list_lca2):
    mset1 = defaultdict(int)
    mset2 = defaultdict(int)

    for lca in list_lca1:
        mset1[",".join(str(x) for x in lca)] += 1

    for lca in list_lca2:
        mset2[",".join(str(x) for x in lca)] += 1

    card = 0

    for k in mset1:
        if k in mset2:
            card += min(mset1[k], mset2[k])

    return card


def is_equal_struct(triple, LCA1, LCA2):
    t = sorted(triple)
    t_set = set(triple)

    triples_nodes_T1 = []

    for node1 in LCA1.label_to_node(t[0]):
        for node2 in LCA1.label_to_node(t[1]):
            for node3 in LCA1.label_to_node(t[2]):
                triples_nodes_T1.append([node1, node2, node3])

    triples_nodes_T2 = []

    for node1 in LCA2.label_to_node(t[0]):
        for node2 in LCA2.label_to_node(t[1]):
            for node3 in LCA2.label_to_node(t[2]):
                triples_nodes_T2.append([node1, node2, node3])

    cmp_vecs_t1 = [[None, None, None] for x in range(len(triples_nodes_T1))]
    ext_t1 = [ExtValue() for x in range(len(triples_nodes_T1))]

    cmp_vecs_t2 = [[None, None, None] for x in range(len(triples_nodes_T2))]
    ext_t2 = [ExtValue() for x in range(len(triples_nodes_T2))]

    for ix_c, couple in enumerate(combinations(t, 2)):
        ix_lb1 = t.index(couple[0])
        ix_lb2 = t.index(couple[1])
        ix_other = list({0, 1, 2} - {ix_lb1, ix_lb2})[0]

        for ix_t1, triple_T1 in enumerate(triples_nodes_T1):
            lca_nd = LCA1.lca_nodes(triple_T1[ix_lb1], triple_T1[ix_lb2])
            lca_lb = LCA1.node_to_labels(lca_nd)
            if len(lca_lb & t_set) > 0:
                cmp_vecs_t1[ix_t1][ix_c] = lca_lb & t_set
            else:
                lca_nd_other = LCA1.lca_nodes(lca_nd, triple_T1[ix_other])
                lca_lb_other = LCA1.node_to_labels(lca_nd_other)
                if len(lca_lb_other & t_set) > 0:
                    cmp_vecs_t1[ix_t1][ix_c] = lca_lb_other & t_set
                else:
                    cmp_vecs_t1[ix_t1][ix_c] = ext_t1[ix_t1].get(lca_nd)

        for ix_t2, triple_T2 in enumerate(triples_nodes_T2):
            lca_nd = LCA2.lca_nodes(triple_T2[ix_lb1], triple_T2[ix_lb2])
            lca_lb = LCA2.node_to_labels(lca_nd)
            if len(lca_lb & t_set) > 0:
                cmp_vecs_t2[ix_t2][ix_c] = lca_lb & t_set
            else:
                lca_nd_other = LCA2.lca_nodes(lca_nd, triple_T2[ix_other])
                lca_lb_other = LCA2.node_to_labels(lca_nd_other)
                if len(lca_lb_other & t_set) > 0:
                    cmp_vecs_t2[ix_t2][ix_c] = lca_lb_other & t_set
                else:
                    cmp_vecs_t2[ix_t2][ix_c] = ext_t2[ix_t2].get(lca_nd)

    missing = True if len(cmp_vecs_t1) == 0 or len(cmp_vecs_t2) == 0 else False

    return (
        missing,
        intersect_mset_card(cmp_vecs_t1, cmp_vecs_t2),
        max(len(cmp_vecs_t1), len(cmp_vecs_t2)),
    )


def get_nset_sig(x_i, x_u):
    return x_u + sigmoid(x_i) * (x_i - x_u)


def similarity(tree1, tree2, mode="sigmoid", sigmoid_mult=10.0, cores=1):
    """
    Compute the similarity score of the two trees.

    Parameters:
    tree1 (Tree): MP3-treesim tree representation
    tree2 (Tree): MP3-treesim tree representation

    Keyword arguments:
    mode (str): 'sigmoid', 'intersection', 'union'
                 or 'geometric',
                 sets the similarity calculation.
                 By default is set to 'sigmoid'.
    sigmoid_mult (float): Multiplicator for the
                 sigmoid calculation.
    cores (int); Number of cores used for the computation.

    Returns:
    float: Similarity score
    """
    if mode not in ["sigmoid", "intersection", "union", "geometric"]:
        raise AttributeError("Incorrect value of mode passed.")

    numerator = 0
    denominator_i = 0
    denominator_u = 0

    if len(set(tree1.label_set) & set(tree2.label_set)) == 0:
        return 0.0

    if mode == "intersection":
        labels = set(tree1.label_set) & set(tree2.label_set)
    else:
        labels = set(tree1.label_set) | set(tree2.label_set)

    combs = combinations(labels, 3)

    if cores <= 0:
        cores = None

    if cores == 1:
        for triple in combs:
            missing, num, dem = is_equal_struct(triple, tree1.LCA, tree2.LCA)
            numerator += num
            if missing:
                denominator_u += dem
            else:
                denominator_i += dem
    else:
        try:
            set_start_method("spawn")
        except RuntimeError:
            print("ERROR")

        with Pool(processes=cores) as pool:
            func = partial(is_equal_struct, LCA1=tree1.LCA, LCA2=tree2.LCA)
            results = pool.map(func, combs)

            for res in results:
                missing, num, dem = res
                numerator += num
                if missing:
                    denominator_u += dem
                else:
                    denominator_i += dem

    similarity_score = 0
    if mode == "intersection":
        similarity_score = float(numerator) / denominator_i
    elif mode == "union":
        similarity_score = float(numerator) / (denominator_i + denominator_u)
    else:
        similarity_score_i = float(numerator) / denominator_i
        similarity_score_u = float(numerator) / (denominator_i + denominator_u)

        if mode == "sigmoid":
            similarity_score = similarity_score_u + sigmoid(
                similarity_score_i, mult=sigmoid_mult
            ) * min(similarity_score_i - similarity_score_u, similarity_score_u)

        elif mode == "geometric":
            similarity_score = np.sqrt(similarity_score_i * similarity_score_u)

    return similarity_score


def build_tree(T, labeled_only=False, exclude=None):
    """
    Build the MP3-treesim tree representation from a networkx representation.

    The tree must have a attribute `label` for each node. Labels in a node
    must be separated by a comma.

    Parameters:
    T (nx.nx_agraph): Tree in networkx representation
    labeled_only (bool): If true nodes without attribute `label`
    will be ignored, meaning that T is a partially labeled tree.
    exclude (list(str)): List of labels to exclude from computation

    Returns:
    Tree: MP3-treesim tree representation
    """
    if not nx.is_tree(T):
        raise ValueError("Not a valid tree.")

    label_to_nodes = defaultdict(set)
    label_set = set()
    node_to_labels = defaultdict(set)

    for node in T.nodes(data=True):
        id_node = node[0]

        if "label" not in node[1] and not labeled_only:
            node[1]["label"] = str(node[0])
        if "label" not in node[1] and labeled_only:
            node[1]["label"] = ""
            continue

        labels = node[1]["label"].split(",")
        for label in labels:
            if exclude:
                if label in exclude:
                    continue

            label_set.add(label)
            label_to_nodes[label].add(id_node)
            node_to_labels[id_node].add(label)

    label_set = list(label_set)
    label_set = sorted(label_set)

    return Tree(T, label_to_nodes, node_to_labels, label_set)



"""Caller Module."""

from collections import OrderedDict

import click

from scphylo.commands.caller._1rename import rename
from scphylo.commands.caller._1sra import sra
from scphylo.commands.caller._2fastp import fastp
from scphylo.commands.caller._3bwa import bwa
from scphylo.commands.caller._3star import star
from scphylo.commands.caller._4gatk import gatk
from scphylo.commands.caller._4rsem import rsem
from scphylo.commands.caller._5hapcaller import hapcaller
from scphylo.commands.caller._5mpileup import mpileup
from scphylo.commands.caller._5mutect2 import mutect2
from scphylo.commands.caller._5sequenza import sequenza
from scphylo.commands.caller._5strelka2 import strelka2
from scphylo.commands.caller._5varscan2 import varscan2
from scphylo.commands.caller._6bamreadcount import bamreadcount
from scphylo.commands.caller._6bamsnap import bamsnap
from scphylo.commands.caller._6defuse import defuse
from scphylo.commands.caller._6pyclone import pyclone
from scphylo.commands.caller._6snpeff import snpeff
from scphylo.commands.caller._6varfilter import varfilter
from scphylo.commands.caller._6vartrix import vartrix


class NaturalOrderGroup(click.Group):
    """Command group trying to list subcommands in the order they were added.

    Make sure you initialize the `self.commands` with OrderedDict instance.
    With decorator, use::
        @click.group(cls=NaturalOrderGroup, commands=OrderedDict())
    """

    def list_commands(self, ctx):
        """List command names as they are in commands dict.

        If the dict is OrderedDict, it will preserve the order commands
        were added.
        """
        return self.commands.keys()


@click.group(
    name="caller",
    cls=NaturalOrderGroup,
    commands=OrderedDict(),
    context_settings={"max_content_width": 300, "terminal_width": 300},
)
def cli_caller():
    r"""Caller module.

    FastQ: {sample}_1.fastq.gz or {sample}.fastq.gz\n
    BAM: {sample}.transcript.bam for counting\n
    BAM: {sample}.mapped.bam -> {sample}.markdup.bam\n
    """
    return None


cli_caller.add_command(rename)
cli_caller.add_command(sra)
cli_caller.add_command(fastp)
cli_caller.add_command(bwa)
cli_caller.add_command(star)
cli_caller.add_command(gatk)
cli_caller.add_command(hapcaller)
cli_caller.add_command(mutect2)
cli_caller.add_command(strelka2)
cli_caller.add_command(varscan2)
cli_caller.add_command(mpileup)
cli_caller.add_command(sequenza)
cli_caller.add_command(varfilter)
cli_caller.add_command(snpeff)
cli_caller.add_command(rsem)
cli_caller.add_command(bamreadcount)
cli_caller.add_command(bamsnap)
cli_caller.add_command(defuse)
cli_caller.add_command(vartrix)
cli_caller.add_command(pyclone)



import click

import scphylo as scp


@click.command(short_help="Boost available tree reconstruction tool (Trisicell-Boost).")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--solver",
    default="scite",
    type=click.Choice(["scite", "phiscs", "scistree"]),
    show_default=True,
    help="Solver of the booster.",
)
@click.option(
    "--sample_on",
    default="muts",
    type=click.Choice(["muts", "cells"]),
    show_default=True,
    help="Sampling on `muts` or `cells`.",
)
@click.option(
    "--sample_size",
    default=10,
    type=int,
    show_default=True,
    help="The size of samples i.e. the number of muts or cells.",
)
@click.option(
    "--n_samples",
    default=10,
    type=int,
    show_default=True,
    help="The number of subsamples.",
)
@click.option(
    "--begin_index",
    default=0,
    type=int,
    show_default=True,
    help="ID of the start subsample name.",
)
@click.option(
    "--n_jobs",
    default=0,
    type=int,
    show_default=True,
    help="Number of parallel jobs to do subsampleing.",
)
@click.option(
    "--dep_weight",
    default=50,
    type=int,
    show_default=True,
    help="""Weight for how many subsamples to be used in dependencies calculation.""",
)
@click.option(
    "--time_limit",
    default=120,
    type=int,
    show_default=True,
    help="Timeout of solving allowance (in second).",
)
@click.option(
    "--n_iterations",
    default=500000,
    type=int,
    show_default=True,
    help="SCITE number of iterations.",
)
@click.option(
    "--subsample_dir",
    default=None,
    type=str,
    show_default=True,
    help="A path to the subsamples directory.",
)
@click.option(
    "--disable_tqdm",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Disable showing the tqdm progress.",
)
@click.option(
    "--no_subsampling",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="No running of subsampling (step 1/3).",
)
@click.option(
    "--no_dependencies",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="No running of subsampling (step 2/3).",
)
@click.option(
    "--no_reconstruction",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="No running of reconstruction (step 3/3).",
)
def booster(
    genotype_file,
    alpha,
    beta,
    solver,
    sample_on,
    sample_size,
    n_samples,
    begin_index,
    n_jobs,
    dep_weight,
    time_limit,
    n_iterations,
    subsample_dir,
    disable_tqdm,
    no_subsampling,
    no_dependencies,
    no_reconstruction,
):
    """Boost available tree reconstruction tool.

    For doing all 3 steps:

    scphylo solver booster input.SC 0.001 0.1
    --solver scite --n_samples 200 --sample_size 15
    --n_jobs 4 --n_iterations 10000 --dep_weight 50
    --subsample_dir . --begin_index 0

    For doing only the last step:

    scphylo solver booster input.SC 0.001 0.1
    --dep_weight 50 --subsample_dir PATH_TO_SUBSAMPLES_FOLDER
    --no_subsampling --no_dependencies
    """
    dirbase = scp.ul.dirbase(genotype_file)
    # dirname, basename = scp.ul.dir_base(genotype_file)

    scp.settings.verbosity = "info"
    if not no_reconstruction:
        scp.settings.logfile = f"{dirbase}.booster.log"

    df_in = scp.io.read(genotype_file)
    df_out = scp.tl.booster(
        df_in,
        alpha=alpha,
        beta=beta,
        solver=solver,
        sample_on=sample_on,
        sample_size=sample_size,
        n_samples=n_samples,
        begin_index=begin_index,
        n_jobs=n_jobs,
        dep_weight=dep_weight,
        time_limit=time_limit,
        n_iterations=n_iterations,
        subsample_dir=subsample_dir,
        disable_tqdm=disable_tqdm,
        no_subsampling=no_subsampling,
        no_dependencies=no_dependencies,
        no_reconstruction=no_reconstruction,
    )
    if not no_reconstruction:
        scp.io.write(df_out, f"{dirbase}.booster.CFMatrix")

    return None



#!/usr/bin/env python
import sys

import anndata as ad
import numpy as np
import pandas as pd
from cyvcf2 import VCF

import scphylo as scp

infile = sys.argv[1]
vcf = VCF(infile)
info = vcf.get_header_type("ANN")["Description"].split(" | ")
info[0] = "Allele"
info[-1] = "ERRORS / WARNINGS / INFO"
info = ["CHROM", "POS", "REF", "ALT", "START", "END"] + info

cells = vcf.samples
muts = []
m_gen = []
m_ref = []
m_alt = []
m_cna = []
for var in VCF(infile):
    if var.FILTER is None:
        if var.is_snp or var.is_indel:
            row = [var.CHROM, var.POS, var.REF, var.ALT, var.start + 1, var.end]
            if var.INFO.get("ANN") is not None:
                ann = var.INFO.get("ANN").split(",")[0].split("|")
                row += ann
            m_gen.append(var.gt_types.tolist())
            m_ref.append(var.gt_ref_depths.tolist())
            m_alt.append(var.gt_alt_depths.tolist())
            muts.append(row)
            cnas_per_var = []

m_gen = np.array(m_gen)
m_ref = np.array(m_ref)
m_alt = np.array(m_alt)
muts = pd.DataFrame(muts, columns=info)
muts.index = muts.index.map(lambda x: f"mut{x}")
cells = pd.DataFrame(index=cells)

adata = ad.AnnData(np.zeros((len(muts), len(cells))))
adata.obs = muts
adata.var = cells
if m_gen.shape[1] > 0:
    adata.layers["genotype"] = m_gen
adata.layers["total"] = m_ref + m_alt
adata.layers["mutant"] = m_alt
adata = adata.T
outdir, basename = scp.ul.dir_base(infile)

bad = [
    "Annotation_Impact",
    "Gene_ID",
    "Feature_ID",
    "Rank",
    "cDNA.pos / cDNA.length",
    "CDS.pos / CDS.length",
    "AA.pos / AA.length",
    "Distance",
    "ERRORS / WARNINGS / INFO",
]
adata.var.drop(bad, axis=1, inplace=True)
if len(adata.obs_names) == 2:
    for obs in adata.obs_names:
        adata.var[f"{obs}_TOTAL_READS"] = adata[obs].layers["total"][0]
        adata.var[f"{obs}_MUTANT_READS"] = adata[obs].layers["mutant"][0]
adata.var.to_csv(outdir + f"/{basename}.tsv", sep="\t")
adata.write(outdir + f"/_{basename[:-len('.ann')]}.h5ad.gz", compression="gzip")



import os

import click

import scphylo as scp
from scphylo.ul._servers import cmd


@click.command(short_help="Run bamsnap.")
# @click.argument(
#     "outdir",
#     required=True,
#     type=click.Path(
#         exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
#     ),
# )
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.argument(
    "bam",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "pos",
    required=True,
    type=str,
)
def bamsnap(ref, bam, pos):
    """Run bamsnap.

    scphylo caller bamsnap ...

    BAM files (*.markdup_bqsr.bam) --> PNG file (_bamsnap.png)
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    cmds = ""
    # scp.ul.mkdir(f"{outdir}/_bamsnap")
    cmds += cmd(
        [
            "bamsnap",
            f"-ref {config['ref']}",
            f"-bam {bam}",
            f"-pos {pos}",
            f"-out ./{pos}",
            "-silence",
            "-read_color_deletion 'C8C8C8'",
            "-coverage_vaf 0.1",
        ],
        islast=True,
    )
    print(cmds)
    # cmds += cmd(["rm -rf", f"{1}_bamsnap.log"])

    os.system(cmds)

    return None



#!/usr/bin/env python
import glob
import sys

import anndata as ad
import pandas as pd


def get_nucleotides(line):
    ref = line.split()[2].upper()
    char = {"A": 0, "C": 0, "G": 0, "T": 0, "N": 0}
    char[ref] = int(line.split()[4].split(":")[1])
    char[line.split()[5].split(":")[0]] = int(line.split()[5].split(":")[1])
    char[line.split()[6].split(":")[0]] = int(line.split()[6].split(":")[1])
    char[line.split()[7].split(":")[0]] = int(line.split()[7].split(":")[1])
    char[line.split()[8].split(":")[0]] = int(line.split()[8].split(":")[1])
    cov = sum(char[a] for a in ["A", "C", "G", "T", "N"])
    return ref, char, cov


def is_mut(x):
    g = 3
    if x["COV"] > 0:
        g = 0
    if x[x["Allele"][0]] > 0:
        g = 1
    return g, x[x["Allele"][0]]


outdir = sys.argv[1]
vcf = pd.read_csv(sys.argv[2])
data = []
for file in glob.glob(f"{outdir}/*.bamreadcount"):
    with open(file) as fin:
        name = file.split("/")[-1].replace(".bamreadcount", "")
        for line in fin:
            line = line.strip()
            a = ".".join(line.split("\t")[:2])
            ref, char, cov = get_nucleotides(line)
            res = {"CELL": name, "MUT": a, "REF": ref, "COV": cov}
            for k, v in char.items():
                res[k] = v
            data.append(res)
df = pd.DataFrame(data)

vcf["MUT"] = vcf["CHROM"] + "." + vcf["POS"].astype(str)
df = pd.merge(df, vcf, how="left", left_on="MUT", right_on="MUT")
df["IS_MUT"] = df.apply(lambda x: is_mut(x)[0], axis=1).astype(int)
df["MUTANT"] = df.apply(lambda x: is_mut(x)[1], axis=1).astype(int)

sc = df.pivot_table(index="CELL", columns="MUT", values="IS_MUT")
adata = ad.AnnData(sc)
adata.layers["mutant"] = df.pivot_table(index="CELL", columns="MUT", values="MUTANT")
adata.layers["total"] = df.pivot_table(index="CELL", columns="MUT", values="COV")
adata.var = pd.merge(adata.var, vcf, how="left", left_index=True, right_on="MUT")
adata.write(outdir + "/_bamreadcount.h5ad.gz", compression="gzip")



import numpy as np
import pandas as pd

import scphylo as scp


def remove_mut_by_list(adata, alist):
    """Remove a list of mutations from the data.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    alist : :obj:`list`
        A list of mutations.
    """
    adata._inplace_subset_var(np.setdiff1d(adata.var_names, alist))
    scp.logg.info(f"Matrix with n_obs × n_vars = {adata.shape[0]} × {adata.shape[1]}")


def keep_mut_by_list(adata, alist):
    """Keep a list of mutations from the data.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    alist : :obj:`list`
        A list of mutations.
    """
    adata._inplace_subset_var(np.intersect1d(alist, adata.var_names))
    scp.logg.info(f"Matrix with n_obs × n_vars = {adata.shape[0]} × {adata.shape[1]}")


def remove_cell_by_list(adata, alist):
    """Remove a list of cells from the data.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    alist : :obj:`list`
        A list of cells.
    """
    adata._inplace_subset_obs(np.setdiff1d(adata.obs_names, alist))
    scp.logg.info(f"Matrix with n_obs × n_vars = {adata.shape[0]} × {adata.shape[1]}")


def keep_cell_by_list(adata, alist):
    """Keep a list of cells from the data.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    alist : :obj:`list`
        A list of mutations.
    """
    adata._inplace_subset_obs(np.intersect1d(alist, adata.obs_names))
    scp.logg.info(f"Matrix with n_obs × n_vars = {adata.shape[0]} × {adata.shape[1]}")


def filter_mut_vaf_greater_than_coverage_mutant_greater_than(
    adata, min_vaf=0.2, min_coverage_mutant=20, min_cells=1
):
    """Remove mutations based on the VAF and coverage.

    This function removes mutations that don't have coverage
    greater than `min_coverage_mutant` and VAF greater than
    `min_vaf` for at least `min_cells`

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    min_vaf : :obj:`float`, optional
        Minimum VAF, by default 0.2
    min_coverage_mutant : :obj:`int`, optional
        Minimum VAF, by default 20
    min_cells : :obj:`int`, optional
        Minimum VAF, by default 1
    """
    V = adata.to_df(layer="mutant")
    T = adata.to_df(layer="total")
    df = V / T
    good_muts = ((df >= min_vaf) & (V >= min_coverage_mutant)).sum(axis=0) >= min_cells
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def filter_mut_mutant_must_present_in_at_least(adata, min_cells=2):
    """Remove mutations based on the mutant status.

    This function removes mutations that don't have the status of mutant
    in at least `min_cells` cells.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    min_cells : :obj:`int`, optional
        Minimum number of cells, by default 1
    """
    G = adata.layers["genotype"]
    good_muts = ((G == 1) | (G == 3)).sum(axis=0) >= min_cells
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def filter_mut_reference_must_present_in_at_least(adata, min_cells=1):
    """Remove mutations based on the wild-type status.

    This function removes mutations that don't have the status of wild-type
    in at least `min_cells` cells. Note that mutations that are present in all cells
    will not be filtered out by this function.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    min_cells : :obj:`int`, optional
        Minimum number of cells, by default 1
    """
    G = adata.layers["genotype"]
    good_muts = ((G == 0).sum(axis=0) >= min_cells) | (
        ((G == 1) | (G == 3)).sum(axis=0) == adata.shape[0]
    )
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def mut_seperated_by_cell_group(adata, group_key):
    def func(x):
        # a = ((x == 1) | (x == 3)).sum(axis=0) / (x != 2).sum(axis=0)
        # b = (x == 2).sum(axis=0) / x.shape[0]
        return ((x == 1) | (x == 3)).sum(axis=0) / x.shape[0]

    grouped = adata.obs.groupby(group_key)
    out = pd.DataFrame(
        np.zeros((adata.shape[1], len(grouped)), dtype=np.float64),
        columns=list(grouped.groups.keys()),
        index=adata.var_names,
    )
    for group, idx in grouped.indices.items():
        X = adata.layers["genotype"][idx]
        out[group] = np.ravel(func(X))
    return out


def remove_nonsense_cells(adata):
    G = adata.layers["genotype"]
    good_cells = ((G == 0) | (G == 2)).sum(axis=1) != G.shape[0]
    keep_cell_by_list(adata, adata.obs_names.to_numpy()[good_cells])


def build_scmatrix(adata):
    G = adata.layers["genotype"]
    adata.X[G == 0] = 0
    adata.X[(G == 1) | (G == 3)] = 1
    adata.X[G == 2] = 3
    adata.X = adata.X.astype("int")

    # M = adata.layers["mutant"]
    # T = adata.layers["total"]
    # adata.X[T != -1] = 3
    # adata.X[T > 0] = 0
    # adata.X[M > 0] = 1
    # adata.X = adata.X.astype("int")


def statistics(adata):
    G = adata.layers["genotype"]
    t = adata.shape[0] * adata.shape[1]
    a = (G == 0).sum().sum()
    b = (G == 1).sum().sum()
    c = (G == 2).sum().sum()
    d = (G == 3).sum().sum()
    scp.logg.info(f"size = {adata.shape[0]} × {adata.shape[1]}")
    scp.logg.info(f"    HOM_REF = {a:6d} ({100*a/t:2.1f}%)")
    scp.logg.info(f"    HET     = {b:6d} ({100*b/t:2.1f}%)")
    scp.logg.info(f"    HOM_ALT = {d:6d} ({100*d/t:2.1f}%)")
    scp.logg.info(f"    UNKNOWN = {c:6d} ({100*c/t:2.1f}%)")


def group_obs_apply_func(adata, group_key, func=np.nansum, layer=None):
    def getX(x):
        if layer is not None:
            return x.layers[layer]
        else:
            return x.X

    grouped = adata.obs.groupby(group_key)
    out = pd.DataFrame(
        np.zeros((adata.shape[1], len(grouped)), dtype=np.float64),
        columns=list(grouped.groups.keys()),
        index=adata.var_names,
    )

    for group, idx in grouped.indices.items():
        X = getX(adata[idx])
        out[group] = np.ravel(func(X, axis=0))
    return out


def filter_snpeff(adata, exome=False):
    a = adata.var.Transcript_BioType == "protein_coding"
    b = adata.var.Feature_Type == "transcript"
    # c = adata.var.Annotation.isin(["synonymous_variant", "missense_variant"])
    # c = adata.var.Annotation.str.contains("intron_variant")
    c = False
    d = adata.var.ALT.apply(lambda x: False if "," in x else True)
    adata._inplace_subset_var(a & b & ~c & d)
    adata.var.drop(["Feature_Type", "Transcript_BioType"], axis=1, inplace=True)
    if exome:
        # tumor_obs = np.setdiff1d(adata.obs_names, ["NB"])[0]
        tumor_obs = adata.obs_names[adata.obs_names.str.contains("_Tumor")][0]
        adata._inplace_subset_obs(adata.obs_names.str.contains(tumor_obs))
        adata.var["Mutant"] = adata.layers["mutant"][0]
        adata.var["Total"] = adata.layers["total"][0]
        adata.var["VAF"] = adata.var["Mutant"] / adata.var["Total"]
        adata.var["SAMPLE"] = tumor_obs



"""Pre/Post-Processing Module."""

from scphylo.pp._bifiltering import bifiltering
from scphylo.pp._binary import (
    binarym_filter_clonal_mutations,
    binarym_filter_nonsense_mutations,
    binarym_filter_private_mutations,
    binarym_statistics,
    consensus_combine,
)
from scphylo.pp._readcount import (
    build_scmatrix,
    filter_mut_mutant_must_present_in_at_least,
    filter_mut_reference_must_present_in_at_least,
    filter_mut_vaf_greater_than_coverage_mutant_greater_than,
    filter_snpeff,
    group_obs_apply_func,
    keep_cell_by_list,
    keep_mut_by_list,
    mut_seperated_by_cell_group,
    remove_cell_by_list,
    remove_mut_by_list,
    remove_nonsense_cells,
    statistics,
)
from scphylo.pp._tree import collapse, sample_from_tree

__all__ = (
    bifiltering,
    binarym_filter_clonal_mutations,
    binarym_filter_nonsense_mutations,
    binarym_filter_private_mutations,
    binarym_statistics,
    consensus_combine,
    build_scmatrix,
    filter_mut_mutant_must_present_in_at_least,
    filter_mut_reference_must_present_in_at_least,
    filter_mut_vaf_greater_than_coverage_mutant_greater_than,
    filter_snpeff,
    keep_cell_by_list,
    keep_mut_by_list,
    remove_cell_by_list,
    remove_mut_by_list,
    statistics,
    collapse,
    group_obs_apply_func,
    sample_from_tree,
    mut_seperated_by_cell_group,
    remove_nonsense_cells,
)



from collections import Counter

import numpy as np
import pandas as pd

import scphylo as scp


def sbm(data2):
    # TODO: implement
    graph_tool, graph_tool_is_not_imported = scp.ul.import_graph_tool()
    if graph_tool_is_not_imported:
        scp.logg.error("Unable to import a package!")

    from graph_tool.inference import minimize as gt_min

    def get_graph(df):
        G = graph_tool.Graph(directed=False)
        vtype = G.new_vertex_property("short")
        label2id = {}

        x = np.where(df == 1)
        for k in range(x[0].shape[0]):
            i = x[0][k]
            j = x[1][k]
            cell_key = f"cell{i}"
            mut_key = f"mut{j}"
            if cell_key in label2id:
                v = label2id[cell_key]
            else:
                v = G.add_vertex()
                label2id[cell_key] = int(v)
                vtype[v] = 1

            if mut_key in label2id:
                w = label2id[mut_key]
            else:
                w = G.add_vertex()
                label2id[mut_key] = int(w)
                vtype[w] = 2

            G.add_edge(v, w)
        return G, label2id, vtype

    def blockmodel_to_labels(b, label2id, n_cells=None, n_muts=None, maxblocks=100):
        if n_cells is None:
            n_cells = max(
                int(x[4:]) if x.startswith("cell") else 0 for x in label2id.keys()
            )
        if n_muts is None:
            n_muts = max(
                int(x[3:]) if x.startswith("mut") else 0 for x in label2id.keys()
            )

        cell_array = [
            (b[label2id[f"cell{i}"]] if f"cell{i}" in label2id else maxblocks + 1)
            for i in range(n_cells)
        ]
        temp = sorted(Counter(cell_array).keys())
        cell_idx_to_blocknum = {temp[i]: i + 1 for i in range(len(temp))}
        cell_array = [cell_idx_to_blocknum[a] for a in cell_array]

        mut_array = [
            (b[label2id[f"mut{i}"]] if f"mut{i}" in label2id else maxblocks + 1)
            for i in range(n_muts)
        ]
        temp = sorted(Counter(mut_array).keys())
        mut_idx_to_blocknum = {temp[i]: i + 1 for i in range(len(temp))}
        mut_array = [mut_idx_to_blocknum[a] for a in mut_array]
        return cell_array, mut_array

    def get_ratio(df):
        x = (df == 1).sum().sum() / df.size
        return x, 1 - x

    data = data2.copy()
    data[data == 3] = 0
    G, label2id, vtype = get_graph(data)

    min_blocks = 2
    max_blocks = 10
    r = gt_min.minimize_nested_blockmodel_dl(
        G, B_min=min_blocks, B_max=max_blocks, state_args={"clabel": vtype}
    )
    b = r.get_bs()[0]

    cell_array, mut_array = blockmodel_to_labels(
        b, label2id, n_cells=data.shape[0], n_muts=data.shape[1]
    )

    ixgrid = np.ix_(np.argsort(cell_array), np.argsort(mut_array))
    N = pd.DataFrame(
        data.values[ixgrid], index=np.sort(cell_array), columns=np.sort(mut_array)
    )
    I_mtr = data2.loc[
        data2.index[np.argsort(cell_array)], data2.columns[np.argsort(mut_array)]
    ]

    for n in set(N.columns):
        for m in set(N.columns):
            if get_ratio(N.loc[n, m])[0] > 0:
                N.loc[n, m] = 1

    N.index = I_mtr.index
    N.columns = I_mtr.columns
    out = N.loc[data2.index, data2.columns]

    return out



import glob
import os
import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run SRA.")
@click.argument(
    "srafile",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "outdir",
    required=True,
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
)
@click.option(
    "--check",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Check the unfinished jobs on biowulf.",
)
def sra(srafile, outdir, check):
    """Run SRA.

    scphylo caller sra path/to/SraRunTable.csv path/to/output/directory

    Columns name of the csv file: `Library Name`, `Run`, `LibraryLayout`
    """
    scp.ul.mkdir(outdir)

    def cmds(srr_id, name, layout):
        cmds = ""
        if layout == "PAIRED":
            cmds += cmd(
                [
                    "fastq-dump",
                    "--split-files",
                    f"--outdir {outdir}",
                    f"{srr_id}",
                ]
            )
            if srr_id != name:
                cmds += cmd(
                    [
                        "mv",
                        f"{outdir}/{srr_id}_1.fastq",
                        f"{outdir}/{name}_1.fastq",
                    ]
                )
                cmds += cmd(
                    [
                        "mv",
                        f"{outdir}/{srr_id}_2.fastq",
                        f"{outdir}/{name}_2.fastq",
                    ]
                )
            cmds += cmd(
                [
                    "gzip",
                    f"{outdir}/{name}_1.fastq",
                ]
            )
            cmds += cmd(
                [
                    "gzip",
                    f"{outdir}/{name}_2.fastq",
                ]
            )
        elif layout == "SINGLE":
            cmds += cmd(
                [
                    "fastq-dump",
                    f"--outdir {outdir}",
                    f"{srr_id}",
                ]
            )
            if srr_id != name:
                cmds += cmd(
                    [
                        "mv",
                        f"{outdir}/{srr_id}.fastq",
                        f"{outdir}/{name}.fastq",
                    ]
                )
            cmds += cmd(
                [
                    "gzip",
                    f"{outdir}/{name}.fastq",
                ]
            )

        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    if not check:
        df = pd.read_csv(srafile, low_memory=False)
        scp.logg.info(f"There are {df.shape[0]} samples.")

        if df["Library Name"].nunique() != df.shape[0]:
            scp.logg.error(
                "Number of unique `Library Name` and total samples are not equal!"
            )

        df["cmd"] = df.apply(
            lambda x: cmds(x["Run"], x["Library Name"], x["LibraryLayout"]), axis=1
        )
        cmdmain = write_cmds_get_main(
            df,
            "sra",
            "12:00:00",
            "10",
            "python/3.7,sratoolkit/2.10.8",
            1,
            "farid.rsh@gmail.com",
            f"{outdir}/_tmp",
            None,
        )
        os.system(cmdmain)
    else:
        n = 0
        for file in glob.glob(f"{outdir}/_tmp/log/_sra/*.o"):
            out = subprocess.getoutput(f"cat {file}")
            if out.count("\nDone!") != out.count("(") and out.count(
                "\nDone!"
            ) != out.count(")"):
                with open(file) as fin:
                    scp.logg.info(
                        fin.readlines()[1].replace(")", "").replace("(", "").strip()
                    )
                    scp.logg.info("\n")
                    n += 1
        scp.logg.info(f"There are {n} unfinished jobs.")

    return None



import os
import time

import networkx as nx
import numpy as np

import scphylo as scp
from scphylo.external._scite import run_scite


def scite(df_input, alpha, beta, n_iters=90000, n_restarts=3, experiment=False):
    """Solving using SCITE.

    Tree inference for single-cell data :cite:`SCITE`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    n_iters : :obj:`int`, optional
        Number of iterations, by default 90000
    n_restarts : :obj:`int`, optional
        Number of restarts, by default 3
    experiment : :obj:`bool`, optional
        Is in the experiment mode (the log won't be shown), by default False

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    if not experiment:
        scp.logg.info(
            f"running SCITE with alpha={alpha}, beta={beta}, n_iters={n_iters}, "
            f"n_restarts={n_restarts}"
        )

    tmpdir = scp.ul.tmpdirsys(suffix=".scite")

    np.savetxt(
        f"{tmpdir.name}/scite.SC.T", df_input.values.T, delimiter="\t", fmt="%1.0f"
    )
    with open(f"{tmpdir.name}/scite.geneNames", "w") as fout:
        fout.write("\n".join(df_input.columns))

    cmd = [
        "scite",
        "-i",
        f"{tmpdir.name}/scite.SC.T",
        "-names",
        f"{tmpdir.name}/scite.geneNames",
        "-n",
        f"{df_input.shape[1]}",
        "-m",
        f"{df_input.shape[0]}",
        "-ad",
        f"{beta}",
        "-fd",
        f"{alpha}",
        "-r",
        f"{n_restarts}",
        "-e",
        "0.20",
        "-a",
        "-l",
        f"{n_iters}",
        "-o",
        f"{tmpdir.name}/scite.output",
    ]

    s_time = time.time()
    run_scite(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    with open(f"{tmpdir.name}/scite.output_ml0.gv") as fin:
        with open(f"{tmpdir.name}/scite.output_ml0_quoted.gv", "w") as fout:
            for line in fin:
                if " -> " in line:
                    line = line.strip()
                    a = line.split(" -> ")[0]
                    b = line.split(" -> ")[1].replace(";", "")
                    fout.write(f'"{a}" -> "{b}";\n')
                else:
                    fout.write(line)

    detail = {}
    with open(f"{tmpdir.name}/scite.log") as fin:
        for line in fin:
            line = line.strip()
            if "best value for beta:" in line:
                detail["beta"] = float(line.replace("best value for beta:", "").strip())
            if "best log score for tree:" in line:
                detail["score"] = float(
                    line.replace("best log score for tree:", "").strip()
                )

    G = nx.drawing.nx_pydot.read_dot(f"{tmpdir.name}/scite.output_ml0_quoted.gv")
    df_output = df_input.copy()
    df_output[:] = 0
    for i in range(df_output.shape[0]):
        muts = nx.shortest_path(G, source="Root", target=f"s{i}")
        muts.remove("Root")
        muts.remove(f"s{i}")
        if len(muts) > 0:
            df_output.loc[df_output.index[i], muts] = 1

    tmpdir.cleanup()

    if not experiment:
        scp.ul.stat(df_input, df_output, alpha, beta, running_time)
        for k, v in detail.items():
            scp.logg.info(f"{k}: {v}")
        return df_output
    else:
        return df_output, running_time, detail["score"], detail["beta"]


def infscite(
    df_input,
    alpha,
    beta,
    n_iters,
    n_restarts=3,
    experiment=False,
):
    # TODO: implement
    scp.logg.info(
        f"running infSCITE with alpha={alpha}, beta={beta}, n_iters={n_iters}, "
        f"n_restarts={n_restarts}"
    )
    tmpdir = scp.ul.tmpdirsys(suffix=".infscite")

    np.savetxt(
        f"{tmpdir.name}/infscite.SC.T", df_input.values.T, delimiter="\t", fmt="%1.0f"
    )
    with open(f"{tmpdir.name}/infscite.geneNames", "w") as fout:
        fout.write("\n".join(df_input.columns))

    infscite = scp.ul.get_file("scphylo.external/bin/infSCITE")
    cmd = (
        f"{infscite} "
        f"-i {tmpdir.name}/infscite.SC.T "
        f"-names {tmpdir.name}/infscite.geneNames "
        f"-n {df_input.shape[1]} "
        f"-m {df_input.shape[0]} "
        f"-ad {beta} "
        f"-fd {alpha} "
        f"-r {n_restarts} "
        "-z "
        "-e 0.20 "
        "-a "
        f"-l {n_iters} "
        f"-o {tmpdir.name}/infscite > {tmpdir.name}/infscite.log"
    )
    # "-rec 1 "
    # "-d "
    # "-s -p 10000 "

    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    with open(f"{tmpdir.name}/infscite_ml0.gv") as fin:
        with open(f"{tmpdir.name}/infscite_ml0_quoted.gv", "w") as fout:
            for line in fin:
                if " -> " in line:
                    line = line.strip()
                    a = line.split(" -> ")[0]
                    b = line.split(" -> ")[1].replace(";", "")
                    fout.write(f'"{a}" -> "{b}";\n')
                else:
                    fout.write(line)

    detail = {}
    with open(f"{tmpdir.name}/infscite.log") as fin:
        for line in fin:
            line = line.strip()
            if "best value for beta:" in line:
                detail["beta"] = float(line.replace("best value for beta:", "").strip())
            if "best value for alpha:" in line:
                detail["alpha"] = float(
                    line.replace("best value for alpha:", "").strip()
                )
            if "best doublet rate:" in line:
                detail["doublet"] = float(
                    line.replace("best doublet rate:", "").strip()
                )
            if "best log score for tree:" in line:
                detail["score"] = float(
                    line.replace("best log score for tree:", "").strip()
                )

    G = nx.drawing.nx_pydot.read_dot(f"{tmpdir.name}/infscite_ml0_quoted.gv")
    df_output = df_input.copy()
    df_output[:] = 0
    for i in range(df_output.shape[0]):
        muts = nx.shortest_path(G, source=f"{df_input.shape[0]}", target=f"{i+1}")
        # muts.remove("Root")
        muts.remove(f"{i+1}")
        if len(muts) > 0:
            df_output.loc[df_output.index[i], muts] = 1

    tmpdir.cleanup()

    if not experiment:
        scp.ul.stat(df_input, df_output, alpha, beta, running_time)
        for k, v in detail.items():
            scp.logg.info(f"{k}: {v}")
        return df_output
    else:
        return df_output, running_time, detail["score"], detail["beta"]



import networkx as nx
import numpy as np
import pandas as pd

import scphylo as scp


def _get_cnt_tree(tree):
    cnt_tree = tree.copy()
    for u, v, _ in cnt_tree.edges.data("label"):
        cnt_tree.add_edge(u, v, label="")
        if cnt_tree.in_degree(v) != 0:
            if "––" not in tree.nodes[v]["label"]:
                cnt_tree.nodes[v]["label"] = cnt_tree.nodes[v]["label"].split(
                    cnt_tree.graph["splitter_cell"]
                )
    return cnt_tree


def _add_artificial_edge(tree):
    artificial_added = []
    cp_tree = tree.copy()
    nodes = list(cp_tree.nodes)
    last_node = max(nodes)
    for n in nodes:
        if cp_tree.in_degree(n) == 1 and cp_tree.out_degree(n) > 0:
            if "––" not in cp_tree.nodes[n]["label"]:
                for x in cp_tree.nodes[n]["label"]:
                    artificial_added.append(x)
                    last_node += 1
                    cp_tree.add_node(last_node, label=[x])
                    cp_tree.add_edge(n, last_node)
                cp_tree.nodes[n]["label"] = "––"
        elif cp_tree.in_degree(n) == 1 and cp_tree.out_degree(n) == 0:
            if len(cp_tree.nodes[n]["label"]) > 1:
                for x in cp_tree.nodes[n]["label"]:
                    artificial_added.append(x)
                    last_node += 1
                    cp_tree.add_node(last_node, label=[x])
                    cp_tree.add_edge(n, last_node)
                cp_tree.nodes[n]["label"] = "––"
    return cp_tree, artificial_added


def _find_set_of_internal_nodes(tree):
    result = []
    for n in tree.nodes:
        if tree.in_degree(n) == 1 and tree.out_degree(n) > 0:
            tmp = []
            nodes = nx.algorithms.traversal.depth_first_search.dfs_tree(tree, n).nodes
            for m in nodes:
                if "––" not in tree.nodes[m]["label"]:
                    tmp += tree.nodes[m]["label"]
            result.append(sorted(tmp))
    tmp = []
    for m in tree.nodes:
        if "––" not in tree.nodes[m]["label"] and tree.in_degree(m) == 1:
            tmp += tree.nodes[m]["label"]
    result.append(sorted(tmp))
    return result


def consensus_day(sc1, sc2):
    good_cells1 = sc1.index[(sc1 != 0).sum(axis=1) > 0]
    good_cells2 = sc2.index[(sc2 != 0).sum(axis=1) > 0]
    common_cells = np.intersect1d(good_cells1, good_cells2)
    if len(common_cells) == 0:
        scp.logg.error("No common cells found in two inputs!")
    sc1_p = sc1.loc[common_cells, :]
    sc2_p = sc2.loc[common_cells, :]

    tree1 = scp.ul.to_tree(sc1_p)
    tree2 = scp.ul.to_tree(sc2_p)

    cnt_tree1 = _get_cnt_tree(tree1)
    cnt_tree2 = _get_cnt_tree(tree2)

    # step 1
    cnt_tree1, artificial_added1 = _add_artificial_edge(cnt_tree1)
    cnt_tree2, artificial_added2 = _add_artificial_edge(cnt_tree2)
    artificial_added = list(np.union1d(artificial_added1, artificial_added2))
    scp.logg.debug(artificial_added)

    # step 2
    soin1 = _find_set_of_internal_nodes(cnt_tree1)
    soin2 = _find_set_of_internal_nodes(cnt_tree2)

    intersection = [x for x in soin1 if x in soin2]
    intersection_len = list(map(len, intersection))
    valmax = max(intersection_len)
    argmax = intersection_len.index(valmax)
    indices = intersection[argmax]
    columns = [f"m{x}" for x in range(len(intersection))]
    df = pd.DataFrame(0, index=indices, columns=columns, dtype=int)
    for i, inter in enumerate(intersection):
        df.loc[inter, f"m{i}"] = 1
    final_tree = scp.ul.to_tree(df)

    # step 3
    final_tree = _get_cnt_tree(final_tree)
    final_tree, _ = _add_artificial_edge(final_tree)

    # step 4
    nodes = list(final_tree.nodes)
    for n in nodes:
        if "––" not in final_tree.nodes[n]["label"] and final_tree.in_degree(n) == 1:
            if final_tree.nodes[n]["label"][0] in artificial_added:
                m = list(final_tree.predecessors(n))[0]
                if "––" in final_tree.nodes[m]["label"]:
                    final_tree.nodes[m]["label"] = final_tree.nodes[n]["label"]
                else:
                    final_tree.nodes[m]["label"] += final_tree.nodes[n]["label"]
                final_tree = nx.contracted_nodes(final_tree, m, n, self_loops=False)

    # convert back to the scphylo tree format
    for v in final_tree.nodes:
        if final_tree.in_degree(v) != 0:
            if "––" not in final_tree.nodes[v]["label"]:
                final_tree.nodes[v]["label"] = final_tree.graph["splitter_cell"].join(
                    final_tree.nodes[v]["label"]
                )
    i = 0
    for e, u, _ in final_tree.edges.data("label"):
        final_tree.edges[(e, u)]["label"] = f"m{i}"
        i += 1
    data = scp.ul.to_cfmatrix(final_tree)
    final_tree = scp.ul.to_tree(data)

    return final_tree



"""scPhylo Module."""

from scphylo import datasets, io
from scphylo import logging as logg
from scphylo import pl, pp, settings, tl, ul

__author__ = ", ".join(["Farid Rashidi"])
__maintainer__ = ", ".join(["Farid Rashidi"])
__email__ = ", ".join(["farid.rsh@gmail.com"])
__version__ = "0.0.5"
__all__ = (datasets, io, logg, pl, pp, settings, tl, ul)



"""Solver Module."""

from scphylo.tl.solver._bnb import bnb
from scphylo.tl.solver._cardelino import cardelino
from scphylo.tl.solver._cellphy import cellphy
from scphylo.tl.solver._dendro import dendro
from scphylo.tl.solver._gpps import gpps
from scphylo.tl.solver._grmt import grmt
from scphylo.tl.solver._onconem import onconem
from scphylo.tl.solver._phiscs import (
    phiscs_readcount,
    phiscsb,
    phiscsb_bulk,
    phiscsi,
    phiscsi_bulk,
)
from scphylo.tl.solver._sbm import sbm
from scphylo.tl.solver._scelestial import scelestial
from scphylo.tl.solver._sciphi import sciphi
from scphylo.tl.solver._scistree import iscistree, rscistree, scistree
from scphylo.tl.solver._scite import infscite, scite
from scphylo.tl.solver._siclonefit import siclonefit
from scphylo.tl.solver._sphyr import sphyr
from scphylo.tl.solver.booster import booster
from scphylo.tl.solver.huntress import huntress

__all__ = (
    bnb,
    cardelino,
    dendro,
    gpps,
    grmt,
    onconem,
    phiscs_readcount,
    phiscsb,
    phiscsb_bulk,
    phiscsi,
    phiscsi_bulk,
    sbm,
    sciphi,
    iscistree,
    rscistree,
    scistree,
    infscite,
    scite,
    siclonefit,
    sphyr,
    booster,
    huntress,
    scelestial,
    cellphy,
)



import os

import click

import scphylo as scp


@click.command(short_help="Run SPhyR.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--n_restarts",
    "-r",
    default=10,
    type=int,
    show_default=True,
    help="Number of restarts.",
)
@click.option(
    "--n_threads",
    "-p",
    default=1,
    type=int,
    show_default=True,
    help="Number of threads.",
)
def sphyr(genotype_file, alpha, beta, n_restarts, n_threads):
    """SPhyR.

    Tumor phylogeny estimation from single-cell sequencing data under loss and error
    :cite:`SPhyR`.

    scphylo solver sphyr input.SC 0.0001 0.1
    """
    outfile = os.path.splitext(genotype_file)[0]

    scp.settings.verbosity = "info"
    scp.settings.logfile = f"{outfile}.sphyr.log"

    df_in = scp.io.read(genotype_file)
    if alpha == 0:
        alpha = 0.000000000001
    df_out = scp.tl.sphyr(
        df_in, alpha=alpha, beta=beta, n_restarts=n_restarts, n_threads=n_threads
    )
    scp.io.write(df_out, f"{outfile}.sphyr.CFMatrix")

    return None



verbosity = 3
#: Verbosity level (0=errors, 1=warnings, 2=infos, 3=hints, debugs=4)

logfile = ""
#: Name of logfile. By default is set to '' and writes to standard output.

tools = ""
#: Path to the directory where external tools are. By default is set to '$PATH'.

refs = ["hg19", "hg38", "mm10", "m10x"]
HUMAN = "/fdb/igenomes/Homo_sapiens/UCSC"
MOUSE = "/fdb/igenomes/Mus_musculus/UCSC"
GATK = "/fdb/GATK_resource_bundle"

hg19 = {}
hg19["ref"] = f"{HUMAN}/hg19/Sequence/WholeGenomeFasta/genome.fa"
hg19["bwa_ref"] = f"{HUMAN}/hg19/Sequence/BWAIndex/genome.fa"
hg19["annot"] = f"{HUMAN}/hg19/Annotation/Genes/genes.gtf"
hg19["known_sites"] = [
    f"{GATK}/hg19/dbsnp_138.hg19.vcf.gz",
    f"{GATK}/hg19/1000G_phase1.snps.high_confidence.hg19.vcf.gz",
    f"{GATK}/hg19/1000G_phase1.indels.hg19.vcf.gz",
    f"{GATK}/hg19/Mills_and_1000G_gold_standard.indels.hg19.vcf.gz",
]
hg19["defuse_config"] = "/usr/local/apps/defuse/0.8.1/config_hg19_ens69.txt"
hg19["defuse_db"] = "/fdb/defuse/hg19_ens69_newgmap"

hg38 = {}
hg38["ref"] = f"{HUMAN}/hg38/Sequence/WholeGenomeFasta/genome.fa"
hg38["bwa_ref"] = f"{HUMAN}/hg38/Sequence/BWAIndex/genome.fa"
hg38["annot"] = f"{HUMAN}/hg38/Annotation/Genes/genes.gtf"
hg38["known_sites"] = [
    f"{GATK}/hg38/dbsnp_138.hg38.vcf.gz",
    f"{GATK}/hg38/1000G_phase1.snps.high_confidence.hg38.vcf.gz",
    f"{GATK}/hg38/Mills_and_1000G_gold_standard.indels.hg38.vcf.gz",
]

mm10 = {}
mm10["ref"] = f"{MOUSE}/mm10/Sequence/WholeGenomeFasta/genome.fa"
mm10["bwa_ref"] = f"{MOUSE}/mm10/Sequence/BWAIndex/genome.fa"
mm10["annot"] = f"{MOUSE}/mm10/Annotation/Genes/genes.gtf"
mm10["known_sites"] = [
    f"{GATK}/mm10/dbsnp146_fixedNames.vcf.gz",
    f"{GATK}/mm10/mgp.v5.merged.indels.dbSNP142.normed.fixedNames.vcf.gz",
    f"{GATK}/mm10/mgp.v5.merged.snps_all.dbSNP142.fixedNames.vcf.gz",
]
mm10["defuse_config"] = "/usr/local/apps/defuse/0.8.1/config_mm10_ens84.txt"
mm10["defuse_db"] = "/fdb/defuse/mm10_ens84_newgmap"

m10x = {}
m10x["ref"] = "/data/rashidimehrabf2/_database/refdata-gex-mm10/fasta/genome.fa"

tools = {}
tools["email"] = "farid.rsh@gmail.com"
tools["bwa"] = "bwa/0.7.17"
tools["samtools"] = "samtools/1.11"
tools["bamtools"] = "bamtools/2.5.1"
tools["bcftools"] = "bcftools/1.9"
tools["fastp"] = "fastp/0.20.1"
tools["star"] = "STAR/2.7.3a"
tools["gatk"] = "GATK/4.2.1.0"
tools["rsem"] = "rsem/1.3.2"
tools["snpeff"] = "snpEff/5.0"
tools["defuse"] = "defuse/0.8.1"
tools["bamreadcount"] = "bamreadcount/0.8.0"
tools["vartrix"] = "vartrix/1.1.22"
tools["strelka"] = "strelka/2.9.10"
tools["varscan"] = "varscan/2.4.3"
tools["sequenza"] = "sequenza-utils/3.0.0"
tools["pyclone"] = "pyclone/0.13.1"



import csv
import math
import os

import numpy as np
from tqdm import tqdm

DIFFERENT_LINEAGES = 0
ANCESTOR_DESCENDANT = 1
DESCENDANT_ANCESTOR = 2
SAME_NODE = 3
UNDEFINED_DEPENDENCY = 4
ROOT_NODE_ID = "-1"  # this is ID of the root node in tree representation
ROOT = ROOT_NODE_ID


def read_matrix_from_file_into_hash(path_matrix_file):
    """Read tab or space delimited file into 2D-dictionary.

    TO DO
    -----
        Add full documentation of the function (not needed at the moment)
    """
    assert os.path.exists(path_matrix_file), (
        "ERROR. There does not exist file " + path_matrix_file
    )

    input_file = open(path_matrix_file)
    input_lines = input_file.readlines()
    input_file.close()

    assert len(input_lines) > 0, "ERROR. Input file " + path_matrix_file + " is empty."

    header_entries = input_lines[0].strip().split()
    column_ids = header_entries[1 : len(header_entries)]
    num_columns = len(column_ids)
    assert num_columns != 0, (
        "ERROR. Number of columns in " + path_matrix_file + " equals zero. Exiting!!!"
    )

    input_lines_without_header = input_lines[1 : len(input_lines)]
    D = {}
    for line in input_lines_without_header:
        line_columns = line.strip().split()
        row_id = line_columns[0].strip()
        assert row_id not in D.keys(), "ERROR. " + row_id + " is already in keys."
        D[row_id] = {}
        for i in range(num_columns):
            assert column_ids[i] not in D[row_id].keys(), (
                "ERROR. " + column_ids[i] + " is already in keys."
            )
            D[row_id][column_ids[i]] = line_columns[i + 1]
    return D


def hash_entries_to_integers(D):
    """Convert all entries of a given dictionary to integers.

    Arguments:
    ---------
    D: dict
        Dictionary of dictionaries (2D-hash)

    Errors:
    ------
        not covered in the current implementation
    """
    for row in D:
        for column in D[row]:
            D[row][column] = int(D[row][column])
    return D


def write_dictionary_of_dictionaries_to_file(D, path_output_file):
    """Write dictionary of dictionaries D to a file.

    The function below can be used to write dictionary of dictionaries D to a file.
    D is in most applications single-cell matrix, e.g. D[cell_id][mut_id] = 1

    Arguments:
    ---------
    D: dictionary of dictionaries
    path_output_file: String (if file exists it will be overwritten)

    """
    row_ids = list(D.keys())
    column_ids = list(D[row_ids[0]].keys())
    assert (
        len(row_ids) > 0 and len(column_ids) > 0
    ), "ERROR. Number of columns or number of rows equal to zero."

    for row_id in row_ids:
        assert set(D[row_id].keys()) == set(column_ids), (
            "ERROR. Some rows have different column ids so this hash is not in a matrix"
            " form,"
        )

    with open(path_output_file, "w") as output_file:
        output_file.write("row_x_col" + "\t" + "\t".join(column_ids) + "\n")
        for row_id in row_ids:
            output_file.write(
                row_id
                + "\t"
                + "\t".join([str(D[row_id][column_id]) for column_id in column_ids])
                + "\n"
            )


def get_row_ids_from_2D_hash(D):
    """Return IDs of rows of 2-dimensional matrix.

    This function returns IDs of rows of 2-dimensional matrix stored via dictionary of
    dictionaries.

    Arguments:
    ---------
    D: dictionary of dictionaries

    Returns:
    -------
    Array of strings representing list of keys of D.
    """
    return list(D.keys())


def get_column_ids_from_2D_hash(D):
    r"""
    Return IDs of columns of 2-dimensional matrix stored via dictionary of dictionaries.

    Arguments:
    ---------
    D: dictionary of dictionaries
        It is assumed that this dictionary is such that D[x].keys() is the same for
        all x \in D.keys().

    Returns:
    -------
    Array of strings representing list of keys of D.
    """
    row_ids = get_row_ids_from_2D_hash(D)
    assert len(row_ids) > 0, "ERROR. Dictionary empty."
    row_0_id = row_ids[0]
    return list(D[row_0_id].keys())


def get_CF_matrix_from_parent_vector(parent, D, alpha, beta):
    """Documentation to be added."""
    cell_ids = list(D.keys())
    mut_ids = list(D[cell_ids[0]].keys())

    children = {}
    children[ROOT] = []
    for mut_id in mut_ids:
        children[mut_id] = []

    for child_id, parent_id in parent.items():
        if child_id != ROOT:
            children[parent_id].append(child_id)

    E = {}
    for cell_id in cell_ids:
        E[cell_id] = {}
        for mut_id in mut_ids:
            E[cell_id][mut_id] = None

    for cell_id in cell_ids:
        score = {}
        score[ROOT] = 0
        for mut_id in mut_ids:
            observed = int(D[cell_id][mut_id])
            if observed == 0:
                score[ROOT] += math.log(1 - alpha)
            if observed == 1:
                score[ROOT] += math.log(alpha)
        best_score = score[ROOT]
        best_mut = ROOT

        muts_to_visit = list(child_id for child_id in children[ROOT])
        while len(muts_to_visit) > 0:
            mut_id = muts_to_visit.pop(0)
            parent_id = parent[mut_id]

            score[mut_id] = score[
                parent_id
            ]  # this is only temporary. see changes below
            observed = int(D[cell_id][mut_id])
            if observed == 0:
                score[mut_id] -= math.log(1 - alpha)
                score[mut_id] += math.log(beta)
            if observed == 1:
                score[mut_id] -= math.log(alpha)
                score[mut_id] += math.log(1 - beta)

            if score[mut_id] > best_score:
                best_score = score[mut_id]
                best_mut = mut_id

            for child_id in children[mut_id]:
                muts_to_visit.append(child_id)

        muts_present_in_true_genotype = []
        current_mut = best_mut
        while current_mut != ROOT:
            muts_present_in_true_genotype.append(current_mut)
            current_mut = parent[current_mut]

        for mut_id in mut_ids:
            if mut_id in muts_present_in_true_genotype:
                E[cell_id][mut_id] = 1
            else:
                E[cell_id][mut_id] = 0

    zero_one_flips = 0
    one_zero_flips = 0

    for cell_id in cell_ids:
        for mut_id in mut_ids:
            observed = int(D[cell_id][mut_id])
            true = int(E[cell_id][mut_id])
            if observed == 1 and true == 0:
                one_zero_flips += 1
            if observed == 0 and true == 1:
                zero_one_flips += 1
    # print("0_1_flips:    " + str(zero_one_flips))
    # print("1_0_flips:    " + str(one_zero_flips))
    return E


def compute_weights_from_dependencies_file(path_dependencies_file):
    """Return weights.

    Arguments:
    ---------
    path_dependencies_file: str
        Path to file which stores dependencies counts obtained after processing folder
        with .CFMatrix files obtained on subsampled instances.
        Expected header of this file can be found below in expected_header_line variable
        and it explains columns present in dependencies file.

    Returns:
    -------
    3-D dictionary weights accessed as follows: weights[mut1][mut2][dependency]
    Here, dependency can be any of ANCESTOR_DESCENDANT, DESCENDANT_ANCESTOR or
    DIFFERENT_LINEAGES.
    Dictionary is not ragged (i.e., both [mut1][mut2] and [mut2][mut1] are populated).
    """
    assert os.path.exists(path_dependencies_file), (
        "ERROR. There does not exist file " + path_dependencies_file
    )

    dependencies_file = open(path_dependencies_file)
    expected_header_line = "MUT1	MUT2	DIFFERENT_LINEAGES	ANCESTOR_DESCENDANT"
    expected_header_line += "	DESCENDANT_ANCESTOR	SAME_NODE	"
    expected_header_line += (
        "UNDEFINED_DEPENDENCY	TOTAL	DOMINANT_DEPENDENCY	PERCENTAGE_DOMINANT"
    )
    observed_header_line = dependencies_file.readline().rstrip()
    assert observed_header_line == expected_header_line, (
        "ERROR for dependencies file "
        + path_dependencies_file
        + ". Header line different from expected."
    )
    dependencies_file.close()

    with open(path_dependencies_file) as dependencies_file:
        dependencies_file_rows = list(csv.DictReader(dependencies_file, delimiter="\t"))

    mut_ids = list({row["MUT1"] for row in dependencies_file_rows})
    # m = len(mut_ids)

    weights = {}
    weights[ROOT] = {}
    for mut_id in mut_ids:
        weights[mut_id] = {}
        weights[ROOT][mut_id] = {ANCESTOR_DESCENDANT: 0, DIFFERENT_LINEAGES: 0}
        weights[mut_id][ROOT] = {ANCESTOR_DESCENDANT: 0, DIFFERENT_LINEAGES: 0}

    for row in dependencies_file_rows:
        mut1 = row["MUT1"]
        mut2 = row["MUT2"]

        weights[mut1][mut2] = {}

        total = (
            int(row["ANCESTOR_DESCENDANT"])
            + int(row["DESCENDANT_ANCESTOR"])
            + int(row["DIFFERENT_LINEAGES"])
            + int(row["SAME_NODE"])
        )

        if total == 0:
            weights[mut1][mut2][DIFFERENT_LINEAGES] = 0
            weights[mut1][mut2][ANCESTOR_DESCENDANT] = 0
            weights[mut1][mut2][DESCENDANT_ANCESTOR] = 0
        else:
            weights[mut1][mut2][DIFFERENT_LINEAGES] = (
                float(row["DIFFERENT_LINEAGES"])
            ) / total
            weights[mut1][mut2][ANCESTOR_DESCENDANT] = (
                float(row["ANCESTOR_DESCENDANT"]) + 0.5 * float(row["SAME_NODE"])
            ) / total
            weights[mut1][mut2][DESCENDANT_ANCESTOR] = (
                float(row["DESCENDANT_ANCESTOR"]) + 0.5 * float(row["SAME_NODE"])
            ) / total

        if row["DOMINANT_DEPENDENCY"] == "UNDEFINED":
            weights[mut1][mut2]["DOMINANT_DEPENDENCY"] = "UNDEFINED"
        else:
            weights[mut1][mut2]["DOMINANT_DEPENDENCY"] = int(row["DOMINANT_DEPENDENCY"])

        weights[mut1][mut2]["TOTAL"] = int(row["TOTAL"])

    return weights


# this function is used to give an order of mutations in which they will be added to
# the big tree
def get_ordered_list_of_mutations_from_input_matrix(path_noisy_SC_matrix):
    D = hash_entries_to_integers(read_matrix_from_file_into_hash(path_noisy_SC_matrix))
    cell_ids = list(D.keys())
    mut_ids = list(D[cell_ids[0]].keys())

    negative_num_ones = (
        {}
    )  # = (-1)*(number of ones in a column). Consequently, mutations having more ones
    # will be added to tree first.
    for mut_id in mut_ids:
        negative_num_ones[mut_id] = (-1) * len(
            [D[cell_id][mut_id] for cell_id in cell_ids if D[cell_id][mut_id] == 1]
        )

    mut_ids.sort(key=lambda mut_id: negative_num_ones[mut_id])

    return mut_ids


def get_next_mutation(weights, existing_mutations):
    mut_ids = list(weights.keys())
    assert len(existing_mutations) < len(
        mut_ids
    ), "ERROR in calling function get_next_mutation. All mutations have been used."

    candidates_for_addition = [x for x in mut_ids if x not in existing_mutations]

    if len(candidates_for_addition) == 1:
        return candidates_for_addition[0]

    else:
        best_score = -1
        best_mutation = None
        for pivot in candidates_for_addition:
            current_score = np.sum(
                [
                    (
                        weights[pivot][mut_id][ANCESTOR_DESCENDANT]
                        - weights[pivot][mut_id][DESCENDANT_ANCESTOR]
                    )
                    for mut_id in candidates_for_addition
                    if mut_id != pivot
                ]
            )
            if current_score > best_score:
                best_score = current_score
                best_mutation = pivot

        return best_mutation


def reconstruct_big_tree(
    path_dependencies_file,
    path_noisy_SC_matrix,
    alpha,
    beta,
    output_files_prefix,
    disable_tqdm,
):

    weights = compute_weights_from_dependencies_file(path_dependencies_file)
    D = hash_entries_to_integers(read_matrix_from_file_into_hash(path_noisy_SC_matrix))

    ordered_mut_ids = get_ordered_list_of_mutations_from_input_matrix(
        path_noisy_SC_matrix
    )  # mutations in order that they should be added to the tree

    A = {}  # A[mut1][mut2] = 1 if and only if mut1 is ancestor of mut2
    A[ROOT] = {}
    A[ROOT][ROOT] = 0
    for mut1 in ordered_mut_ids:
        A[mut1] = {}
        A[ROOT][mut1] = 1
        A[mut1][ROOT] = 0
        for mut2 in ordered_mut_ids:
            if mut2 != mut1:
                A[mut1][mut2] = None

    parent = {}  # parent vector (keys are mutation IDs)
    parent[ROOT] = "NA"
    existing_mutations = [ROOT]  # mutations added so far to the tree
    big_tree_score = 0

    # m = len(ordered_mut_ids)

    for mutation_to_add in tqdm(
        ordered_mut_ids,
        ascii=True,
        ncols=100,
        desc="RECONSTRUCTION (3/3)",
        position=0,
        disable=disable_tqdm,
    ):
        # print(
        #     "Number of mutations added: "
        #     + str(len(existing_mutations))
        #     + " out of "
        #     + str(m)
        # )

        best_choice_move = None
        best_choice_mutation = None
        best_choice_tree_score = -1

        for pivot_mut in existing_mutations:
            current_move = "child"  # add new mutation as child to the pivot mutation
            current_tree_score = big_tree_score
            for mut_id in existing_mutations:
                if (A[mut_id][pivot_mut] == 1) or (mut_id == pivot_mut):
                    current_tree_score += weights[mut_id][mutation_to_add][
                        ANCESTOR_DESCENDANT
                    ]
                else:
                    current_tree_score += weights[mut_id][mutation_to_add][
                        DIFFERENT_LINEAGES
                    ]

            if current_tree_score > best_choice_tree_score:
                best_choice_tree_score = current_tree_score
                best_choice_mutation = pivot_mut
                best_choice_move = current_move

            # edge (parent, pivot_mut) is broken into
            # (parent, pivot_mut) + (pivot_mut, mutation_to_add)
            current_move = "break edge lower"
            current_tree_score = big_tree_score
            for mut_id in existing_mutations:
                if (A[mut_id][pivot_mut] == 1) or (mut_id == pivot_mut):
                    current_tree_score += weights[mut_id][mutation_to_add][
                        ANCESTOR_DESCENDANT
                    ]
                elif A[pivot_mut][mut_id] == 1:
                    current_tree_score += weights[mutation_to_add][mut_id][
                        ANCESTOR_DESCENDANT
                    ]
                else:
                    current_tree_score += weights[mut_id][mutation_to_add][
                        DIFFERENT_LINEAGES
                    ]

            if current_tree_score > best_choice_tree_score:
                best_choice_tree_score = current_tree_score
                best_choice_mutation = pivot_mut
                best_choice_move = current_move

            # edge (parent, pivot_mut) is broken into
            # (parent, mutation_to_add) + (mutation_to_add, pivot_mut)
            current_move = "break edge middle"
            if pivot_mut == ROOT:
                continue
            # we just use previous current score with swapping pivot and mutation_to_add
            current_tree_score -= weights[pivot_mut][mutation_to_add][
                ANCESTOR_DESCENDANT
            ]
            current_tree_score += weights[mutation_to_add][pivot_mut][
                ANCESTOR_DESCENDANT
            ]
            if current_tree_score > best_choice_tree_score:
                best_choice_tree_score = current_tree_score
                best_choice_mutation = pivot_mut
                best_choice_move = current_move

        big_tree_score = best_choice_tree_score
        if best_choice_move == "child":
            parent[mutation_to_add] = best_choice_mutation
            for mut_id in existing_mutations:
                if (
                    A[mut_id][best_choice_mutation] == 1
                    or mut_id == best_choice_mutation
                ):
                    A[mut_id][mutation_to_add] = 1
                    A[mutation_to_add][mut_id] = 0
                else:
                    A[mut_id][mutation_to_add] = 0
                    A[mutation_to_add][mut_id] = 0

        if best_choice_move == "break edge lower":
            for mut_id in existing_mutations:
                if parent[mut_id] == best_choice_mutation:
                    parent[mut_id] = mutation_to_add
            parent[mutation_to_add] = best_choice_mutation

        if best_choice_move == "break edge middle":
            parent[mutation_to_add] = parent[best_choice_mutation]
            parent[best_choice_mutation] = mutation_to_add

        if best_choice_move in ["break edge lower", "break edge middle"]:
            for mut_id in existing_mutations:
                if mut_id == best_choice_mutation:
                    if best_choice_move == "break edge lower":
                        A[mut_id][mutation_to_add] = 1
                        A[mutation_to_add][mut_id] = 0
                    else:
                        A[mut_id][mutation_to_add] = 0
                        A[mutation_to_add][mut_id] = 1
                elif A[best_choice_mutation][mut_id] == 1:
                    A[mutation_to_add][mut_id] = 1
                    A[mut_id][mutation_to_add] = 0
                elif A[mut_id][best_choice_mutation] == 1:
                    A[mut_id][mutation_to_add] = 1
                    A[mutation_to_add][mut_id] = 0
                else:
                    A[mutation_to_add][mut_id] = 0
                    A[mut_id][mutation_to_add] = 0

        existing_mutations.append(mutation_to_add)
        A[mutation_to_add][mutation_to_add] = 0

    with open(output_files_prefix + ".dnc.tree.parent", "w") as parent_vector_file:
        for mut_id, parent_id in parent.items():
            parent_vector_file.write(mut_id + "\t" + parent_id + "\n")

    E = get_CF_matrix_from_parent_vector(parent, D, alpha, beta)
    write_dictionary_of_dictionaries_to_file(E, output_files_prefix + ".dnc.CFMatrix")

    return big_tree_score



import itertools

import ete3
import networkx as nx
import numpy as np

import scphylo as scp
from scphylo.external._mp3 import build_tree, similarity
from scphylo.ul._trees import _to_newick


def bourque(df_grnd, df_sol):
    """Bourque distances for mutation trees.

    This measure was introduced in :cite:`Bourque`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """
    # TODO: implement
    return None


def pcss(df_grnd, df_sol):
    """Pairwise cell shortest-path similarity score.

    For every pair of cells :math:`i` and :math:`j`, we computed the shortest-path
    :math:`d_{ij}` between the two cells in each tree. If the two cells belong to the
    same clone, their shortest-path distance is 0, otherwise the shortest-path distance
    equals the number of edges (regardless of direction) that separate the clones of the
    two cells. Finally, we summed up the absolute differences between the shortest-path
    distances of all unordered pairs of cells in the two trees.
    This measure is metric. The proof is given in the paper.

    This measure was introduced in :cite:`OncoNEM`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """
    # TODO: implement
    return None


def mp3(df_grnd, df_sol):
    """Triplet-based similarity score.

    For fully multilabeled trees with poly-occurring labels.
    This measure was introduced in :cite:`MP3`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """
    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    if len(inter) == 0:
        scp.logg.error("No common mutations found between two trees!")
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)
    inter = np.setdiff1d(
        inter,
        np.union1d(
            tree_grnd.graph["become_germline"], tree_sol.graph["become_germline"]
        ),
    )
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)
    tree_grnd = scp.ul.to_mtree(tree_grnd)
    tree_sol = scp.ul.to_mtree(tree_sol)

    for n in tree_grnd.nodes:
        if tree_grnd.in_degree(n) > 0:
            tree_grnd.nodes[n]["label"] = ",".join(tree_grnd.nodes[n]["label"])
    for n in tree_sol.nodes:
        if tree_sol.in_degree(n) > 0:
            tree_sol.nodes[n]["label"] = ",".join(tree_sol.nodes[n]["label"])

    T1 = build_tree(tree_grnd)
    T2 = build_tree(tree_sol)
    return similarity(T1, T2)


def caset(df_grnd, df_sol):
    """Commonly Ancestor Sets score.

    This measure was introduced in :cite:`CASet_DISC`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """

    def _get_ancesteral_set(tree):
        root = scp.ul.root_id(tree)
        ancesteral_set = {}
        for n in tree.nodes:
            if tree.in_degree(n) > 0 and "––" not in tree.nodes[n]["label"]:
                for mut in tree.nodes[n]["label"]:
                    ancester_set = ["root"]
                    ancester_set += tree.nodes[n]["label"]  # self ancestor is ok
                    for m in nx.shortest_path(tree, root, n):
                        if tree.in_degree(m) > 0 and "––" not in tree.nodes[m]["label"]:
                            ancester_set += tree.nodes[m]["label"]
                    ancesteral_set[mut] = ancester_set
        return ancesteral_set

    def _get_common_ancesteral_set(muts, ancesteral_set):
        common_ancesteral_set = {}
        for x, y in itertools.combinations(muts, 2):
            common_ancesteral_set[(x, y)] = np.intersect1d(
                ancesteral_set[x], ancesteral_set[y]
            )
        return common_ancesteral_set

    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    if len(inter) == 0:
        scp.logg.error("No common mutations found between two trees!")
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)
    inter = np.setdiff1d(
        inter,
        np.union1d(
            tree_grnd.graph["become_germline"], tree_sol.graph["become_germline"]
        ),
    )
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)
    tree_grnd = scp.ul.to_mtree(tree_grnd)
    tree_sol = scp.ul.to_mtree(tree_sol)

    ancesteral_set_grnd = _get_ancesteral_set(tree_grnd)
    ancesteral_set_sol = _get_ancesteral_set(tree_sol)

    common_ancesteral_set_grnd = _get_common_ancesteral_set(inter, ancesteral_set_grnd)
    common_ancesteral_set_sol = _get_common_ancesteral_set(inter, ancesteral_set_sol)

    final = []
    for x, y in itertools.combinations(inter, 2):
        a = len(
            np.intersect1d(
                common_ancesteral_set_grnd[(x, y)], common_ancesteral_set_sol[(x, y)]
            )
        )
        b = len(
            np.union1d(
                common_ancesteral_set_grnd[(x, y)], common_ancesteral_set_sol[(x, y)]
            )
        )
        final.append(a / b)

    return np.mean(final)


def disc(df_grnd, df_sol):
    """Distinctly Inherited Sets score.

    This measure was introduced in :cite:`CASet_DISC`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """

    def _get_ancesteral_set(tree):
        root = scp.ul.root_id(tree)
        ancesteral_set = {}
        for n in tree.nodes:
            if tree.in_degree(n) > 0 and "––" not in tree.nodes[n]["label"]:
                for mut in tree.nodes[n]["label"]:
                    ancester_set = ["root"]
                    ancester_set += tree.nodes[n]["label"]  # self ancestor is ok
                    for m in nx.shortest_path(tree, root, n):
                        if tree.in_degree(m) > 0 and "––" not in tree.nodes[m]["label"]:
                            ancester_set += tree.nodes[m]["label"]
                    ancesteral_set[mut] = ancester_set
        return ancesteral_set

    def _get_distinctly_inherited_set(muts, ancesteral_set):
        common_ancesteral_set = {}
        for x, y in itertools.permutations(muts, 2):
            common_ancesteral_set[(x, y)] = np.setdiff1d(
                ancesteral_set[x], ancesteral_set[y]
            )
        return common_ancesteral_set

    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    if len(inter) == 0:
        scp.logg.error("No common mutations found between two trees!")
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)
    inter = np.setdiff1d(
        inter,
        np.union1d(
            tree_grnd.graph["become_germline"], tree_sol.graph["become_germline"]
        ),
    )
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)
    tree_grnd = scp.ul.to_mtree(tree_grnd)
    tree_sol = scp.ul.to_mtree(tree_sol)

    ancesteral_set_grnd = _get_ancesteral_set(tree_grnd)
    ancesteral_set_sol = _get_ancesteral_set(tree_sol)

    distinctly_inherited_set_grnd = _get_distinctly_inherited_set(
        inter, ancesteral_set_grnd
    )
    distinctly_inherited_set_sol = _get_distinctly_inherited_set(
        inter, ancesteral_set_sol
    )

    final = []
    for x, y in itertools.combinations(inter, 2):
        a = len(
            np.intersect1d(
                distinctly_inherited_set_grnd[(x, y)],
                distinctly_inherited_set_sol[(x, y)],
            )
        )
        b = len(
            np.union1d(
                distinctly_inherited_set_grnd[(x, y)],
                distinctly_inherited_set_sol[(x, y)],
            )
        )
        if (
            b > 0
        ):  # FIXME: if a and b are in the same node distinctly_inherited_set is empty
            final.append(a / b)

    return np.mean(final)


def rf(df_grnd, df_sol):
    """Robinson-Foulds score.

    The Robinson–Foulds or symmetric difference metric is defined as (A + B) where A is
    the number of partitions of data implied by the first tree but not the second tree
    and B is the number of partitions of data implied by the second tree but not the
    first tree (although some software implementations divide the RF metric by 2
    and others scale the RF distance to have a maximum value of 1).

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        Similarity out of one.
    """
    inter = np.intersect1d(df_grnd.index, df_sol.index)
    if len(inter) == 0:
        scp.logg.error("No common cells found between two trees!")
    df_grnd1 = df_grnd.loc[inter]
    df_sol1 = df_sol.loc[inter]

    tree_grnd = scp.ul.to_tree(df_grnd1)
    tree_sol = scp.ul.to_tree(df_sol1)

    nwk_grnd = _to_newick(tree_grnd)
    nwk_sol = _to_newick(tree_sol)

    # from skbio import TreeNode
    # tree_grnd = TreeNode.read([nwk_grnd])
    # tree_sol = TreeNode.read([nwk_sol])
    # return tree_grnd.compare_rfd(tree_sol)

    tree_grnd = ete3.Tree(nwk_grnd, format=1)
    tree_sol = ete3.Tree(nwk_sol, format=1)

    rf = tree_grnd.robinson_foulds(tree_sol, unrooted_trees=True)
    return 1 - rf[0] / rf[1]



import glob
import os

import click

import scphylo as scp


@click.command(short_help="Run Rename.")
@click.argument(
    "indir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "keep",
    required=True,
    type=str,
)
@click.option(
    "--exec",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Execute the renaming.",
)
def rename(indir, keep, exec):
    """Run Rename.

    scphylo caller rename path/to/in/dir 0,1,3 --exec
    """
    if keep == ",":
        pass
    else:
        keep = list(map(int, keep.split(",")))
    idx = 0
    cell_i = 1
    for file1 in sorted(glob.glob(indir + "/*.fastq.gz")):
        dirname, basename = scp.ul.dir_base(file1)
        if keep == ",":
            file2 = f"cell{cell_i}"
            if idx % 2 == 1:
                cell_i += 1
        else:
            file2 = [x for i, x in enumerate(basename.split("_")) if i in keep]
            file2 = "_".join(file2)
        if idx % 2 == 0:
            file2 += "_1.fastq.gz"
        else:
            file2 += "_2.fastq.gz"
        file2 = dirname + "/" + file2
        if not exec:
            scp.logg.info(file1, file2)
        else:
            os.rename(file1, file2)
        idx += 1
    return None



import pandas as pd
from IPython.display import Image, display

import scphylo as scp


def dendro(
    adata,
    cell_info=None,
    label_color="group_color",
    layout="fan",
    tiplab_size=2.5,
    width=800,
    height=800,
    dpi=100,
):
    """DENDRO.

    Genetic heterogeneity profiling and subclone detection by single-cell
    RNA sequencing :cite:`DENDRO`.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        Input data contains layers of `genotype`, `mutant` and `total`.
    cell_info : :class:`pandas.DataFrame`, optional
        Information about cells such as color, by default None
    label_color : :obj:`str`, optional
        The column name in which colors of cells are stored
        in the dataframe provided as `cell_info`, by default "group_color"
    tiplab_size : :obj:`float`, optional
        Cell name size in the tree, by default 2.5
    layout : :obj:`str`, optional
        Layout of the tree, by default "fan"
        Values are:

            - `fan`
            - `dendrogram`
    width : :obj:`int`, optional
        Width of the figure, by default 800
    height : :obj:`int`, optional
        Height of the figure, by default 800
    dpi : :obj:`int`, optional
        The resolution, by default 100

    Returns
    -------
    :obj:`None`
    """
    dendro, dendro_is_not_imported = scp.ul.import_rpy2(
        "DENDRO", "devtools::install_github('zhouzilu/DENDRO')\n"
    )
    if dendro_is_not_imported:
        scp.logg.error("Unable to import a package!")
    ggtree, ggtree_is_not_imported = scp.ul.import_rpy2(
        "ggtree",
        "devtools::install_github(c('YuLab-SMU/ggtree','xiangpin/ggtreeExtra'"
        + ",'YuLab-SMU/aplot'))\ninstall.packages('cowplot')\n",
    )
    if ggtree_is_not_imported:
        scp.logg.error("Unable to import a package!")

    import rpy2.robjects as ro
    from rpy2.robjects import pandas2ri
    from rpy2.robjects.lib import grdevices
    from rpy2.robjects.packages import importr

    stats = importr("stats")

    scp.logg.info("running DENDRO")

    Z = adata.to_df(layer="genotype")
    Z[(Z == 1) | (Z == 3)] = 1
    Z[Z == 2] = 0
    X = adata.to_df(layer="mutant")
    N = adata.to_df(layer="total")

    with ro.conversion.localconverter(ro.default_converter + pandas2ri.converter):
        Z = ro.conversion.py2rpy(Z.T)
        X = ro.conversion.py2rpy(X.T)
        N = ro.conversion.py2rpy(N.T)

        if cell_info is not None:
            cell_info_r = ro.conversion.py2rpy(cell_info.reset_index())
        else:
            cell_info_r = ro.conversion.py2rpy(
                pd.DataFrame.from_dict(
                    {
                        "x": adata.obs.index,
                        "group_color": adata.obs.shape[0] * ["#000000"],
                    }
                )
            )
        ro.globalenv["info"] = cell_info_r

    dist = dendro.DENDRO_dist(X, N, Z, show_progress=False)
    tree = stats.hclust(dist, method="ward.D")
    ro.globalenv["tree"] = tree

    cmd = f"""
    p <- ggtree(tree, layout='{layout}', size=0.4) %<+% info
    p <- p + geom_tiplab(
        aes(color={label_color}), align=TRUE, linesize=.1, angle=-90,
        hjust=-0.1, size={tiplab_size}, fontface='bold'
    )
    p <- p + scale_color_identity(guide='none')
    """

    with grdevices.render_to_bytesio(
        grdevices.png, width=width, height=height, res=dpi
    ) as image:
        p = ro.r(cmd)
        ro.r.show(p)

    return display(Image(image.getvalue(), embed=True, retina=True))



import random
from functools import lru_cache

import numpy as np

import scphylo as scp
from scphylo.external.gpps._utils_hc import (
    copy_tree,
    import_ilp_out,
    prune_and_reattach,
)

__author__ = "Simone Ciccolella"
__date__ = "11/30/21"


@lru_cache(maxsize=None)
def cell_row_likelihood(input_row, node_genotype, alpha, beta):
    likelihood = 0
    for j in range(len(input_row)):
        if input_row[j] == "0":
            if node_genotype[j] == "0":
                likelihood += np.log(1 - beta)
            elif node_genotype[j] == "1":
                likelihood += np.log(alpha)
            else:
                likelihood += -np.inf
        elif input_row[j] == "1":
            if node_genotype[j] == "0":
                likelihood += np.log(beta)
            elif node_genotype[j] == "1":
                likelihood += np.log(1 - alpha)
            else:
                likelihood += -np.inf

        elif input_row[j] == "2":
            likelihood += 0
    return likelihood


def greedy_tree_likelihood(tree, nid_dict, input_scs, alpha, beta):
    type(tree)
    likelihood = 0
    attachment = []
    for row in input_scs:
        str_row = "".join(str(int(x)) for x in row)
        best_lh = -np.inf
        best_attachment = -1
        for node in nid_dict:
            str_gt = "".join(str(int(x)) for x in nid_dict[node].genotype_profile)

            lh = cell_row_likelihood(str_row, str_gt, alpha, beta)
            if lh > best_lh:
                best_lh = lh
                best_attachment = node
        likelihood += best_lh
        attachment.append(best_attachment)

    return likelihood, attachment


def get_expect_matrix(tree, nid_dict, input_scs, alpha, beta):
    _, attachment = greedy_tree_likelihood(tree, nid_dict, input_scs, alpha, beta)
    e_matrix = []

    for node in nid_dict:
        t_node = nid_dict[node]
        if (
            t_node.loss
            and t_node.id_node not in attachment
            and len(t_node.children) == 0
        ):
            scp.logg.debug(t_node.name)

    for c in range(len(attachment)):
        e_matrix.append(nid_dict[attachment[c]].genotype_profile)
    return e_matrix


def generate_neighborhood(start_tree, start_nid_dict, neighborhood_size):
    type(start_nid_dict)
    neighbors = []
    while len(neighbors) < neighborhood_size:
        # prune-reattach only
        cp_tree, cp_dict = copy_tree(start_tree)
        node_ids = list(cp_dict.keys())
        prune = random.choice(node_ids)
        reattach = random.choice(node_ids)

        if prune != reattach:
            if prune_and_reattach(cp_dict[prune], cp_dict[reattach], cp_dict):
                neighbors.append((cp_tree, cp_dict))
            else:
                cp_tree = None
                cp_dict = None
    return neighbors


def hill_climbing(
    start_tree,
    start_nid_dict,
    neighborhood_size,
    max_iterations,
    alpha,
    beta,
    input_scs,
):
    current_tree = start_tree
    current_dict = start_nid_dict
    current_lh, _ = greedy_tree_likelihood(
        start_tree, start_nid_dict, input_scs, alpha, beta
    )
    scp.logg.debug("Initial log-likelihood: %f" % current_lh)

    current_iteration = 1
    while current_iteration < max_iterations:

        if current_iteration % 10 == 0:
            scp.logg.debug("Current iteration: %d" % current_iteration)

        neighbors = generate_neighborhood(current_tree, current_dict, neighborhood_size)
        next_eval = -np.inf
        next_sol = None

        for ng in neighbors:
            # scp.logg.debug(ng)
            ng_lh, _ = greedy_tree_likelihood(ng[0], ng[1], input_scs, alpha, beta)

            if ng_lh > next_eval:
                next_eval = ng_lh
                next_sol = (ng[0], ng[1])

        if next_eval > current_lh:
            current_tree = next_sol[0]
            current_dict = next_sol[1]
            current_lh = next_eval
            scp.logg.debug("Found a better solution with likelihood: %f" % current_lh)
        current_iteration += 1

    return current_tree, current_dict


def gpps_hc(input_matrix, ilp_matrix, alpha, beta, k_dollo, mut_names, ns=30, mi=100):
    imported_tree, imported_nid_dict = import_ilp_out(ilp_matrix, k_dollo, mut_names)

    hc_best_tree, hc_best_dict = hill_climbing(
        imported_tree,
        imported_nid_dict,
        neighborhood_size=ns,
        max_iterations=mi,
        alpha=alpha,
        beta=beta,
        input_scs=input_matrix,
    )

    e_mat = get_expect_matrix(hc_best_tree, hc_best_dict, input_matrix, alpha, beta)

    return e_mat



import subprocess

import click

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run MuTect2.")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "normal",
    required=True,
    type=str,
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default="0-10:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="50",
    type=str,
    show_default=True,
    help="Memory.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def mutect2(outdir, normal, ref, time, mem, afterok):
    """Run MuTect2.

    scphylo caller mutect2 /path/to/in/dir normal_name hg19|hg38|mm10

    BAM files (*.markdup_bqsr.bam) --> VCF files (*.mutect2.vcf)
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def get_command(sample):
        cmds = ""
        cmds += cmd([f"module load {scp.settings.tools['gatk']}"])
        cmds += cmd(
            [
                f'gatk --java-options "-Xmx{int(mem)-10}g"',
                "Mutect2",
                f"--reference {config['ref']}",
                f"--input {outdir}/{sample}.markdup_bqsr.bam",
                f"--input {outdir}/{normal}.markdup_bqsr.bam",
                f"--normal-sample {normal}",
                f"--output {outdir}/{sample}.tmp.mutect2.vcf",
            ]
        )
        cmds += cmd(
            [
                f'gatk --java-options "-Xmx{int(mem)-10}g"',
                "FilterMutectCalls",
                f"--reference {config['ref']}",
                f"--variant {outdir}/{sample}.tmp.mutect2.vcf",
                f"--output {outdir}/{sample}.mutect2.vcf",
            ]
        )
        cmds += cmd(
            [
                "rm -rf",
                f"{outdir}/{sample}.mutect2.vcf.filteringStats.tsv",
                f"{outdir}/{sample}.mutect2.vcf.stats",
                f"{outdir}/{sample}.mutect2.vcf.idx",
                f"{outdir}/{sample}.tmp.mutect2.vcf",
                f"{outdir}/{sample}.tmp.mutect2.vcf.stats",
                f"{outdir}/{sample}.tmp.mutect2.vcf.idx",
            ]
        )
        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    df_cmds = scp.ul.get_samples_df(outdir, normal)
    df_cmds["cmd"] = df_cmds.apply(lambda x: get_command(x["sample"]), axis=1)

    cmdmain = write_cmds_get_main(
        df_cmds,
        "mutect2",
        time,
        mem,
        None,
        1,
        scp.settings.tools["email"],
        f"{outdir}/_tmp",
        afterok,
    )
    code = subprocess.getoutput(cmdmain)
    scp.logg.info(code)

    return None



import os

import click
import numpy as np
from joblib import Parallel, delayed

import scphylo as scp


@click.command(short_help="Run SCITE.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--n_iters",
    "-l",
    default=1000000,
    type=int,
    show_default=True,
    help="Number of iterations.",
)
@click.option(
    "--n_restarts",
    "-r",
    default=3,
    type=int,
    show_default=True,
    help="Number of restarts.",
)
@click.option(
    "--experiment",
    "-e",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Is in experiment mode.",
)
@click.option(
    "--time_limit",
    "-t",
    default=86400,
    type=float,
    show_default=True,
    help="Time limit for the experiment part (in seconds).",
)
@click.option(
    "--smooth_rate",
    "-s",
    default=2,
    type=float,
    show_default=True,
    help="Smooth rate for the experiment part.",
)
@click.option(
    "--iters_rate",
    "-i",
    default=30000,
    type=int,
    show_default=True,
    help="Interations rate for the experiment part.",
)
def scite(
    genotype_file,
    alpha,
    beta,
    n_iters,
    n_restarts,
    experiment,
    time_limit,
    smooth_rate,
    iters_rate,
):
    """SCITE.

    Tree inference for single-cell data :cite:`SCITE`.

    scphylo solver scite input.SC 0.0001 0.1 -l 1000000 -r 3 -e -t 86400 -s 2
    """
    outfile = os.path.splitext(genotype_file)[0]

    scp.settings.verbosity = "info"

    df_in = scp.io.read(genotype_file)
    if not experiment:
        scp.settings.logfile = f"{outfile}.scite.log"
        df_out = scp.tl.scite(
            df_in,
            alpha=alpha,
            beta=beta,
            n_iters=n_iters,
            n_restarts=n_restarts,
        )
        scp.io.write(df_out, f"{outfile}.scite.CFMatrix")
    else:
        scp.settings.logfile = f"{outfile}.scite.log"
        df_out, running_time, _, _ = scp.tl.scite(
            df_in,
            alpha=alpha,
            beta=beta,
            n_iters=iters_rate,
            n_restarts=1,
            experiment=True,
        )
        n_iters = int(smooth_rate * iters_rate * time_limit / running_time)

        def run(i):
            do, r, s, b = scp.tl.scite(
                df_in,
                alpha=alpha,
                beta=beta,
                n_iters=n_iters,
                n_restarts=1,
                experiment=True,
            )
            return do, r, s, b

        output = Parallel(n_jobs=n_restarts)(delayed(run)(i) for i in range(n_restarts))

        scores = [x[2] for x in output]
        betas = [x[3] for x in output]
        best_i = np.argmax(scores)
        df_out = output[best_i][0]

        scp.ul.stat(df_in, df_out, alpha, beta, output[best_i][1])
        scp.logg.info(f"score: {output[best_i][2]}")
        scp.logg.info(f"beta: {output[best_i][3]}")
        scp.logg.info(f"n_iters: {n_iters}")
        scp.logg.info(f"scores: {','.join(list(map(str, scores)))}")
        scp.logg.info(f"betas: {','.join(list(map(str, betas)))}")
        scp.logg.info(f"picked: {best_i}")

        scp.io.write(df_out, f"{outfile}.scite.CFMatrix")

    return None



"""Setup Module."""

import os
import sys
from pathlib import Path
from sys import platform

from setuptools import find_packages, setup
from setuptools.extension import Extension

if platform == "linux" or platform == "linux2":
    os.environ["CC"] = "g++"
elif platform == "darwin":
    os.environ["CC"] = "clang++"

extensions = [
    Extension(
        "scphylo.external._mltd",
        sources=[
            "src/scphylo/external/_mltd.pyx",
            "src/scphylo/external/mltd/mltd.cpp",
        ],
        include_dirs=["src/scphylo/external/mltd"],
        extra_compile_args=["-std=c++11"],
        language="c++",
    ),
    Extension(
        "scphylo.external._scprob",
        sources=[
            "src/scphylo/external/_scprob.pyx",
            "src/scphylo/external/scprob/main.cpp",
        ],
        include_dirs=["src/scphylo/external/scprob"],
        extra_compile_args=["-std=c++11"],
        language="c++",
    ),
    Extension(
        "scphylo.external._scistree",
        sources=[
            "src/scphylo/external/_scistree.pyx",
            "src/scphylo/external/scistree/Utils.cpp",
            "src/scphylo/external/scistree/Utils2.cpp",
            "src/scphylo/external/scistree/Utils3.cpp",
            "src/scphylo/external/scistree/Utils4.cpp",
            "src/scphylo/external/scistree/UtilsNumerical.cpp",
            "src/scphylo/external/scistree/RerootTreeUtils.cpp",
            "src/scphylo/external/scistree/TreeBuilder.cpp",
            "src/scphylo/external/scistree/UnWeightedGraph.cpp",
            "src/scphylo/external/scistree/MarginalTree.cpp",
            "src/scphylo/external/scistree/RBT.cpp",
            "src/scphylo/external/scistree/PhylogenyTreeBasic.cpp",
            "src/scphylo/external/scistree/PhylogenyTree.cpp",
            "src/scphylo/external/scistree/BioSequenceMatrix.cpp",
            "src/scphylo/external/scistree/BinaryMatrix.cpp",
            "src/scphylo/external/scistree/GenotypeMatrix.cpp",
            "src/scphylo/external/scistree/ScistGenotype.cpp",
            "src/scphylo/external/scistree/ScistPerfPhyUtils.cpp",
            "src/scphylo/external/scistree/ScistPerfPhyImp.cpp",
            "src/scphylo/external/scistree/ScistDoublet.cpp",
            "src/scphylo/external/scistree/ScistErrRateInf.cpp",
            "src/scphylo/external/scistree/main.cpp",
        ],
        include_dirs=["src/scphylo/external/scistree"],
        extra_compile_args=["-O3", "-std=c++11", "-c"],
        language="c++",
    ),
    Extension(
        "scphylo.external._scite",
        sources=[
            "src/scphylo/external/_scite.pyx",
            "src/scphylo/external/scite/matrices.cpp",
            "src/scphylo/external/scite/mcmcBinTreeMove.cpp",
            "src/scphylo/external/scite/mcmc.cpp",
            "src/scphylo/external/scite/mcmcTreeMove.cpp",
            "src/scphylo/external/scite/output.cpp",
            "src/scphylo/external/scite/rand.cpp",
            "src/scphylo/external/scite/scoreBinTree.cpp",
            "src/scphylo/external/scite/scoreTree.cpp",
            "src/scphylo/external/scite/treelist.cpp",
            "src/scphylo/external/scite/trees.cpp",
            "src/scphylo/external/scite/findBestTrees.cpp",
        ],
        include_dirs=["src/scphylo/external/scite"],
        extra_compile_args=["-O3", "-std=c++11", "-c"],
        language="c++",
    ),
]


def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                sfile = path + ".cpp"
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


try:
    from Cython.Build import cythonize

    HAS_CYTHON = True
except ImportError:
    HAS_CYTHON = False

CYTHONIZE = bool(int(os.getenv("CYTHONIZE", 1 if HAS_CYTHON else 0)))
if CYTHONIZE:
    if not HAS_CYTHON:
        sys.stderr.write(
            "Cannot find Cython. Have you installed all the requirements?\n"
            "Try pip install Cython\n"
        )
        sys.exit(1)
    compiler_directives = {"language_level": 2, "embedsignature": True}
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)

if __name__ == "__main__":
    setup(
        name="scphylo-tools",
        version="0.0.5",
        ext_modules=extensions,
        description=Path("README.rst").read_text("utf-8").split("\n")[3],
        package_dir={"": "src"},
        packages=find_packages(where="src"),
    )
